#!/usr/bin/env python

from magpy.stream import *
from magpy.core.methods import is_number
from magpy.opt import cred as mpcred
from martas.core import methods as mm
from martas import collector as mcoll
from martas.version import __version__
import dash
from dash import html, dash_table, dcc, Output, Input, State, callback
import dash_daq as daq
import dash_bootstrap_components as dbc
from plotly.subplots import make_subplots
from martas.app.monitor import _latestfile
from crontab import CronTab
import paho.mqtt
import paho.mqtt.client as mqtt
import threading
import socket
import ssl


"""
Required packages
dash
plotly
pip install dash_daq
"""

class MartasPage(object):
    def __init__(self, martasconfig=None, livedata=None):
        """
        DESCRIPTION
        The general martas class which is initiated when starting the martas page
        """
        print (" IMPORTANT: will not run correctly with debug = True on main")
        print ("          : as debug will run flask twice which also starts two conflicting mqtt clients")
        self.debug = False
        if not livedata:
            livedata = {}
        if not martasconfig:
            mainpath = os.path.dirname(os.path.realpath(__file__))
            configpath = os.path.join(mainpath, "..", "..", "conf", "martas.cfg")
            webconfigpath = os.path.join(mainpath, "..", "..", "conf", "web.cfg")
        else:
            configpath = martasconfig
            webconfigpath = martasconfig
            self.debug = True

        cfg = mm.get_conf(configpath)
        webcfg = mm.get_conf(webconfigpath)
        self.logpath = os.path.dirname(cfg.get('logging'))
        self.sensorpath = cfg.get('sensorsconf')
        self.bufferpath = cfg.get('bufferdirectory')
        # import a webpage config: darkmode, defaultpage, default limits, displayrate, read file first
        self.read_initial_buffer = True
        self.defaultpage = None  # set default # get from config
        if webcfg.get('defaultpage','martas') in ['martas','Martas','MARTAS','homedirectory']:
            self.defaultpage = "/" # set default # get from config
        self.srate = int(webcfg.get('refreshrate',5)) # displayrate - needs to be large enough, dynamically adjusted

        self.defaultheight = int(webcfg.get('defaultheight', 450))
        self.defaultduration = int(webcfg.get('defaultcoverage', 10))
        if self.defaultduration > 60:
            self.defaultduration = 60
        self.data2show = self.get_new_data()
        self.cfg = cfg
        self.livedata = livedata
        self.statusdict = {"mqtt" : {"space" : 400, "used": 150 },
                      "base" : {"space" : 100, "used": 100 }
                      }

        if webcfg.get('read_initial_buffer',False) in ['False','false','FALSE', False]:
            self.read_initial_buffer = False
        if webcfg.get('debug',False) in ['True','true','TRUE', True]:
            self.debug = True
        if self.read_initial_buffer:
            print("Using buffer to get initial data")
            for d in self.data2show:
                data, names, anames, sr, sensorid = self.extract_data(d,duration=self.defaultduration*60)
                self.livedata[sensorid] = {'samplingrate': sr, 'names': names, 'allnames': anames, 'data': data}
        print ("Display refresh rate: {} sec".format(self.srate))
        # now create initial outputs:
        self.sl = self.get_sensors(debug=False)
        self.sensors_to_plot = self.get_sensors(debug=False) # will be updated eventually


        self.t1_stop = threading.Event()
        self.martas_live_monitor(debug=self.debug)

    def get_sensors(self, debug=False):
        """
        DESCRIPTION
            read SensorIDs from sensors
        """
        sl = []
        path = self.sensorpath
        sensorlistmain = mm.get_sensors(path, "")
        sensorlist = mm.get_sensors(path, "!")
        if len(sensorlist) > 0:
            # drop OW group input from main list if OW sensors found
            sensorlistmain = [el for el in sensorlistmain if not el.get("sensorid","").find("OW") > -1]
        for line in sensorlist:
            sl.append(line.get('sensorid',''))
        sensorlist = mm.get_sensors(path, "?")
        if len(sensorlist) > 0:
            sensorlistmain = [el for el in sensorlistmain if not el.get("sensorid","").find("ARDUINO") > -1]
        for line in sensorlist:
            sl.append(line.get('sensorid',''))
        sensorlist = mm.get_sensors(path, "$")
        if len(sensorlist) > 0:
            sensorlistmain = [el for el in sensorlistmain if not el.get("name","").find("MySQL") > -1]
        for line in sensorlist:
            sl.append(line.get('sensorid',''))
        for line in sensorlistmain:
            sl.append(line.get('sensorid',''))
        if debug:
            print ("Activated sensors in sensors.cfg", sl)
        return sl

    def get_new_data(self, duration=600, debug=False):
        """
        DESCRIPTION
            Get all current data files from mqtt path
        """
        path = self.bufferpath
        data2show = []
        dirs=[x[0] for x in os.walk(path)]
        if debug:
            print ("Found the following directories:", dirs)
        for d in dirs:
            ld = _latestfile(os.path.join(d,'*'),date=True)
            lf = _latestfile(os.path.join(d,'*'))
            if os.path.isfile(lf):
                # check white and blacklists
                performtest = False
                now = datetime.now(timezone.utc).replace(tzinfo=None)
                diff = (now-ld).total_seconds()
                if debug:
                    print ("File and time diff from now:", lf, diff, duration)
                if diff < duration/2.:
                    data2show.append(lf)
        return data2show

    def get_sensor_table(self, duration=600, debug=False):
        """
        DESCRIPTION
            Get all current data files from mqtt path
        """
        sl = self.sl
        scols = ['SensorID','Components','Active']
        stable = []
        if len(sl) > 0:
            for s in sl:
                options=[]
                values=[]
                #htmllist.append(html.Span('{}'.format(s), style=style))
                active = False
                for d in sl:
                    if d.find(s) >= 0:
                        active = True
                        if self.livedata.get(d):
                            ndd = self.livedata.get(d)
                            for elem in ndd.get('allnames',''):
                                values.append(str(elem))
                stable.append({'SensorID' : s, 'Components':",".join(values),'Active':str(active)})
        return stable, scols

    def getspace(self, path="/srv"): # path = '/srv'
        """
        DESCRIPTION
            get space from monitor
        """
        statvfs = os.statvfs(path)
        total = (statvfs.f_frsize * statvfs.f_blocks / (1024.*1024.))     # Size of filesystem in bytes
        remain = (statvfs.f_frsize * statvfs.f_bavail / (1024.*1024.))     # Number of free bytes that ordinary users
        usedper=100-(remain/total*100.)
        used = total-remain
        return total, used

    def get_storage_usage(self, debug=False):
        statusdict = self.statusdict
        txt = ""
        atotal, aused = 0, 0
        mtotal, mused = 0, 0
        if self.bufferpath:
            if os.path.exists(self.bufferpath):
                atotal, aused = self.getspace(path=self.bufferpath)
        if self.logpath:
            logpath = os.path.dirname(self.logpath)
            if os.path.exists(logpath):
                mtotal, mused = self.getspace(path=logpath)
        ad = statusdict.get('mqtt',{})
        if ad:
            ad["space"] = np.round(atotal/1000.,0)
            ad["used"] = np.round(aused/1000.,0)
        md = statusdict.get('base',{})
        if md:
            md["space"] = np.round(mtotal/1000.,0)
            md["used"] = np.round(mused/1000.,0)
        if debug:
            print ("SPACE", atotal, aused)
            print("SPACE", mtotal, mused)
        return statusdict


    def get_cron_jobs(self, debug=False):
        """
        DESCRIPTION
            get active Cron jobs once per minute
        """
        crontable = []

        mycron = CronTab(user=True)
        for job in mycron:
            if debug:
                print ("Testing job:", job)
            comm = job.comment
            cand = job.command
            en = job.is_enabled()
            cl = cand.split()
            logstatus = False
            active = False
            jobcommand = cl[1]
            lf = ""
            ctime = ""
            if len(jobcommand) < 3:
                jobcommand = cl[2]
            jc = jobcommand.split("/")[-1]
            for el in cl:
                if el.find(".log") >= 0:
                    te = el.split("/")
                    if debug:
                        print ("Found basic cron job", te)
                    try:
                        ctime = os.path.getctime(el)
                        ctime = datetime.fromtimestamp(ctime)
                        logstatus = True
                        if ctime > datetime.now()-timedelta(days=8):
                            active = True
                        ctime = ctime.strftime("%Y-%m-%dT%H:%M:%S")
                        jn = jc.split(".")[0]
                    except:
                        pass
                    lf = te[-1]
            resd = {"job" : jc, "enabled":str(en), "logfile":lf, "last log":ctime}
            crontable.append(resd)
        croncols = ["job", "enabled", "logfile", "last log"]

        return crontable, croncols

    def extract_data(self, f, duration=600, debug=False):
        if debug:
            print("Extracting data from:", f)
        mindisplayrate = 5
        t1 = datetime.now()
        stream = read(f, starttime=datetime.now(timezone.utc).replace(tzinfo=None) - timedelta(seconds=duration))
        if stream.samplingrate() < 0.99:
            mindisplayrate = 60
            if debug:
                print ("Sampling rate to high for rendering display - downsampling to 1 sec")
            stream = stream.filter()
        # convert to pandas
        keys = stream._get_key_headers()
        if debug:
            print(" - got {} datapoints for keys {}".format(len(stream), keys))
        data = {}
        names = []
        allnames = []
        data['time'] = stream.ndarray[0]
        for key in keys:
            # Limit to numerical keys for selection
            name = stream.get_key_name(key)
            if key in DataStream().NUMKEYLIST and len(names) < 3:
                # limit to three components for each sensor
                data[name] = stream._get_column(key)
                names.append(name)
            allnames.append(name)
        # print (data, names)
        t2 = datetime.now()
        #tdiff = (t2-t1).total_seconds()
        #if tdiff > mindisplayrate/2:
        #    mindisplayrate = int(np.ceil(tdiff*2))
        return data, names, allnames, stream.samplingrate(), stream.header.get("SensorID")


    def live_timer(self, client, stop_event):
            """"
            DESCRIPTION
                timer for live data monitoring
            :param typus:
            :param stop_event:
            :return:
            """

            debug = False
            while(not stop_event.is_set()):
                if debug:
                    # use a green light to indicate it is running
                    print ("Running ... {}".format(datetime.now(timezone.utc).replace(tzinfo=None)))
                stop_event.wait(self.srate)
            ###
            client.loop_stop()


    def martas_live_monitor(self, linelimit=36000, debug=False):
            """
            DEFINITION:
                embbed matplotlib figure in canvas for monitoring

            PARAMETERS:
                linelimit:  store this amount of data within the array: 36000 would contain 1hour in 10Hz/0.1sec resolution
            """
            #global livedata
            values = []
            config = self.cfg

            broker = config.get("broker","")
            martastopic = config.get("station","").lower()
            port = config.get("mqttport","")
            timeout = config.get("mqttdelay","")
            credentials=config.get('mqttcredentials',"").strip()
            qos = config.get("mqttqos","")
            user = config.get("user","")
            password = config.get("password","")
            martascert = config.get("mqttcert","")
            martaspsk = config.get("mqttpsk","")
            dbcred = ""

            if debug:
                print ("Sensors", self.sl, martastopic)

            # start monitoring parameters
            def connectclient(broker='localhost', port=1883, timeout=60, credentials='', user='', password='',
                              qos=0, mqttcert="", mqttpsk="", mqttversion=2, destinationid='', debug=False):
                """
                connectclient method
                used to connect to a specific client as defined by the input variables
                eventually add multiple client -> {"clients":[{"broker":"192.168.178.42","port":"1883"}]} # json type
                        import json
                        altbro = json.loads(altbrocker)
                """
                ## create a unique clientid consisting of broker, client and destination
                client = None
                hostname = socket.gethostname()
                clientid = "{}{}{}".format(broker, hostname, destinationid)

                ## create MQTT client
                ##  ----------------------------
                pahovers = paho.mqtt.__version__
                pahomajor = int(pahovers[0])
                print(" paho-mqtt version ", pahovers)
                try:
                    client = mqtt.Client(clientid, False)
                except:
                    try:
                        client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2, client_id=clientid)
                    except:
                        client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION1, client_id=clientid)

                # TLS encryption part
                if int(port) == 8883 and not mqttpsk:
                    if mqttcert:
                        if debug:
                            print("MQTT: TLS encryption based on certificate")
                            print(mqttcert)
                        client.tls_set(ca_certs=mqttcert)
                    else:
                        if debug:
                            print("MQTT: basic TLS")
                        client.tls_set(ca_certs=None, certfile=None, keyfile=None, cert_reqs=ssl.CERT_REQUIRED,
                                       tls_version=ssl.PROTOCOL_TLS,
                                       ciphers=None)
                if int(port) in [8884, 8883] and mqttpsk:
                    if debug:
                        print("MQTT: TLS encrytion based in PSK")
                    # making use of discussions in https://github.com/eclipse-paho/paho.mqtt.python/issues/451
                    pskidentity = mpcred.lc(mqttpsk, 'user')
                    pskpwd = mpcred.lc(mqttpsk, 'passwd')
                    try:
                        context = mcoll.SSLPSKContext(ssl.PROTOCOL_TLSv1_2)
                        print("MARTAS: Ignore the eventual deprecation warning - it is necessary for old clients")
                        context.set_ciphers('PSK')
                        context.psk = bytes.fromhex(pskpwd)
                        context.identity = pskidentity.encode()
                        client.tls_set_context(context)  # Here we apply the new `SSLPSKContext`
                    except:
                        print("PSK to be included")

                # Authentication part
                if not credentials in ['', '-']:
                    # use user and pwd from credential data if not yet set
                    if user in ['', None, 'None', '-']:
                        user = mpcred.lc(credentials, 'user')
                    if password in ['', '-']:
                        password = mpcred.lc(credentials, 'passwd')
                if not user in ['', None, 'None', '-']:
                    # client.tls_set(tlspath)  # check http://www.steves-internet-guide.com/mosquitto-tls/
                    client.username_pw_set(user,
                                           password=password)  # defined on broker by mosquitto_passwd -c passwordfile user
                client.on_connect = on_connect
                # on message needs: stationid, destination, location
                client.on_message = on_message
                client.connect(broker, port, timeout)

                return client

            # Version 1 (valid for xxx)
            # -------------------------
            # The callback for when the client receives a CONNACK response from the server.
            # signature suitable for MQTT v5.0 client:
            def on_connect(client, userdata, flags, reason_code, properties=None):
                # Subscribing in on_connect() means that if we lose the connection and
                # reconnect then subscriptions will be renewed.
                client.subscribe("{}/#".format(martastopic), qos)

            # The callback for when a PUBLISH message is received from the server.
            def on_message(client, userdata, msg):
                if debug:
                    print(f"Received `{msg.payload.decode()}` from `{msg.topic}` topic")
                sensorid = "{}".format(msg.topic).replace("{}/".format(martastopic),'')[:-5]
                content = "{}".format(msg.topic)[-4:]
                payload = "{}".format(msg.payload.decode())
                if sensorid in self.sl:
                    #sensorcont = {}
                    namelst = []
                    sensorcont = self.livedata.get(sensorid,{})
                    keys = sensorcont.get('SensorKeys', '').split(',')
                    multi = sensorcont.get('Multipliers', '').split(',')
                    els = sensorcont.get('SensorElements','').split(',')
                    uns = sensorcont.get('SensorUnits','').split(',')
                    sr = sensorcont.get("samplingrate",1)
                    pc = sensorcont.get('PackingCode', '')
                    datacont = sensorcont.get('data', {})
                    #print ("Existing datacont", datacont)
                    if content == 'meta':
                        #print ("FOUND META ---------------------------")
                        if not keys or keys == ['']:
                            payloadlist1 = payload.split(sensorid)
                            payloadlist2 = payloadlist1[1].split()
                            sensorcont['SensorKeys'] = payloadlist2[0][1:-1]
                            sensorcont['SensorElements'] = payloadlist2[1][1:-1]
                            sensorcont['SensorUnits'] = payloadlist2[2][1:-1]
                            sensorcont['Multipliers'] = payloadlist2[3][1:-1]
                            sensorcont['PackingCode'] = payloadlist2[4]
                            keys = sensorcont.get('SensorKeys').split(',')
                            els = sensorcont.get('SensorElements').split(',')
                            uns = sensorcont.get('SensorUnits').split(',')
                            for i, el in enumerate(keys):
                                sensorcont[f"col-{el}"] = els[i]
                                sensorcont[f"unit-col-{el}"] = uns[i]
                    elif content == 'dict':
                        payloadlist = payload.split(',')
                        for pl in payloadlist:
                            plc = pl.replace("\n", "").split(":")
                            sensorcont[plc[0]] = plc[1].replace('-', '')
                    elif content == 'data' and len(keys) > 0 and not keys==['']:
                        payloadlist = payload.split(';')
                        for dataline in payloadlist:
                            datalist = dataline.split(',')
                            datalist = [int(el) if is_number(el) else el for el in datalist]
                            #print (datalist, array)
                            timel = [int(t) for t in datalist[:7]]
                            values.append(timel)
                            ti = list(datacont.get('time',[]))
                            ti.append(datetime(*timel))
                            ti = ti[-linelimit:]
                            if len(ti) > 5 and not sensorcont.get("samplingrate",None):
                                # determine sampling rate if not set already from file data - needed for maximum display range
                                sensorcont["samplingrate"] = np.median(np.diff(ti)).total_seconds()
                            datacont['time'] = np.asarray(ti)
                            #print ("Length of time column", len(ti))
                            if pc.endswith('6hL'):
                                sectimel = [int(t) for t in datalist[-7:]]
                                sti = list(datacont.get('sectime',[]))
                                sti.append(datetime(*sectimel))
                                sti = sti[-linelimit:]
                                datacont['sectime'] = np.asarray(sti)
                            for i, k in enumerate(keys):
                                if not k == "sectime" and k in DataStream().KEYLIST:
                                    nam = els[i]
                                    namelst.append(nam)
                                    if k in DataStream().NUMKEYLIST:
                                        dat = list(datacont.get(nam, []))
                                        dat.append(float(datalist[7 + i]) / float(multi[i]))
                                        dat = dat[-linelimit:]
                                        datacont[nam] = np.asarray(dat)
                                    else:
                                        dat = list(datacont.get(nam, []))
                                        dat.append(datalist[7 + i])
                                        dat = dat[-linelimit:]
                                        datacont[nam] = np.asarray(dat)
                        # limit length of array
                        sensorcont['allnames'] = namelst
                        # by default limit to three shown diagrams for each sensor
                        sensorcont['names'] = namelst[:3]
                        sensorcont['data'] = datacont
                    self.livedata[sensorid] = sensorcont
                else:
                    print ("NEW SensorID: ", sensorid)
                    self.sl.append(sensorid)

            mqttversion = int(config.get("mqttversion", 2))
            mqttcert = config.get("mqttcert", "")
            mqttpsk = config.get("mqttpsk", "")
            client = connectclient(broker, port, timeout, credentials, user, password, qos, mqttcert=mqttcert,
                                       mqttpsk=mqttpsk, mqttversion=mqttversion, destinationid=dbcred,
                                       debug=debug)  # dbcred is used for clientid

            if debug:
                print ("Connecting to:", broker, int(port), int(timeout))

            #
            t1 = threading.Thread(target=self.live_timer, args=(client, self.t1_stop))
            t1.start()
            client.loop_start()
            #client.loop_forever()
            # Display the plot

configpath = None
tester = False
if tester:
    configpath = "/home/leon/.martas/conf/martas.cfg"

# Initialize the MARTAS page class and start MQTT recording
mapa = MartasPage(configpath)

statusdict = mapa.get_storage_usage(debug=False)
stable, scols = mapa.get_sensor_table()
ctable, ccols = mapa.get_cron_jobs(debug=False)
cron_columns = [{'id': c, 'name': c} for c in ccols]
sens_columns = [{'id': c, 'name': c} for c in scols]
initially_selected_rows = list(range(0, len(stable)))

dash.register_page(__name__, path=mapa.defaultpage, top_nav=True, assets_ignore='marcos.css')

layout = (html.Div(
                 className="wrapper",
                 children=[
                              html.Div(
                                  className="box martas",
                                  children=html.Div([
                                      html.Img(src='/assets/header.png',style={'width': '30%'}),
                                      ]),
                              ),
                              html.Div(
                                  className="box setup",
                                  children=html.Div([
                                      html.P('Time coverage (minutes):'),
                                      dcc.Slider(id='martas-durationslider', min=5, max=60, step=5, value=mapa.defaultduration,
                                                 marks={x: str(x) for x in [10, 20, 30, 40, 50, 60]}),
                                      html.P('Height of the graph display (px):'),
                                      dcc.Slider(id='martas-heightslider', min=200, max=1000, step=25, value=mapa.defaultheight,
                                                 marks={x: str(x) for x in [200, 400, 600, 800, 1000]}),
                                  ])
                              ),
                              html.Div(
                                  className="box sensors",
                                  children=html.Div(dbc.Collapse([
                                      dcc.Interval(
                                          id='martas-sensors-update',
                                          interval=5*mapa.srate*1000,  # in milliseconds - use 60 sec to allow even LEMI 10Hz data to loaded
                                          n_intervals=0
                                      ),
                                      dash_table.DataTable(id='martas-sensors-table',
                                                            data=stable,
                                                            sort_action='native',
                                                            row_selectable="multi",
                                                            selected_rows=initially_selected_rows,
                                                            columns=sens_columns,
                                                            style_cell={'minWidth': 95, 'width': 95, 'maxWidth': 95,
                                                                        'padding': '5px'},
                                                            style_cell_conditional=[
                                                                {
                                                                    'if': {'column_id': c},
                                                                    'textAlign': 'left'
                                                                } for c in ['SensorID', 'Components']
                                                            ],
                                                            style_data={
                                                                'color': 'rgb(230, 230, 250)',
                                                                'fontSize': '14px',
                                                                'backgroundColor': 'rgb(112, 128, 144)'
                                                            },
                                                            style_data_conditional=[
                                                                {
                                                                    'if': {'row_index': 'odd'},
                                                                    'backgroundColor': 'rgb(119, 149, 163)',
                                                                },
                                                                {
                                                                    'if': {
                                                                        'filter_query': '{Active} = True',
                                                                        'column_id': 'Active'
                                                                    },
                                                                    'backgroundColor': "#008003"
                                                                },
                                                                {
                                                                    'if': {
                                                                        'filter_query': '{Active} = False',
                                                                        'column_id': 'Active'
                                                                    },
                                                                    'backgroundColor': 'tomato'
                                                                }
                                                            ],
                                                            style_header={
                                                                'backgroundColor': '#4D4D4D',
                                                                'color': 'rgb(230, 230, 250)',
                                                                'fontWeight': 'bold',
                                                                'fontSize': '16px',
                                                                'border': '1px solid black'
                                                            }
                                      )
                                  ],
                                  id="horizontal-collapse1",
                                  is_open=True,
                                  dimension="width",
                                  )),
                              ),
                              html.Div(
                                  className="box gauges",
                                  children=html.Div(dbc.Collapse([
                                      dcc.Interval('martas-gauge-update', interval=10*mapa.srate*1000, n_intervals=0),
                                      daq.Gauge(
                                          id='martas-buffer-gauge',
                                          size=120,
                                          color={"gradient": True,
                                                 "ranges": {"green": [0, statusdict['mqtt'].get('space') * 0.8],
                                                            "yellow": [statusdict['mqtt'].get('space') * 0.8,
                                                                       statusdict['mqtt'].get('space') * 0.9],
                                                            "red": [statusdict['mqtt'].get('space') * 0.9,
                                                                    statusdict['mqtt'].get('space')]}},
                                          value=statusdict['mqtt'].get('used'),
                                          label='Buffer [GB]',
                                          max=statusdict['mqtt'].get('space'),
                                          min=0,
                                      )
                                  ],
                                  id="horizontal-collapse2",
                                  is_open=True,
                                  dimension="width",
                                  )),
                              ),
                              html.Div(
                                  id="grid-graphs",
                                  className='box graphs',
                                  children=html.Div([
                                      dcc.Graph(id='martas-live-graph', animate=False),
                                      dcc.Interval(
                                          id='martas-graph-update',
                                          interval=mapa.srate * 1000,  # in milliseconds - use 60 sec to allow even LEMI 10Hz data to loaded
                                          n_intervals=0
                                      )
                                  ])
                              ),
                              html.Div(
                                  className="box link",
                                  children=html.Div([
                                      dcc.Interval('martas-cron-update', interval=10*mapa.srate*1000, n_intervals=0),
                                      dash_table.DataTable(id='martas-cron-table',
                                                           data=ctable,
                                                           columns=cron_columns,
                                                           sort_action='native',
                                                           fixed_rows={'headers': True},
                                                           style_cell={'minWidth': 95, 'width': 95, 'maxWidth': 95,
                                                                       'padding': '5px'},
                                                           style_table={'height': 200},  # default is 500
                                                           style_cell_conditional=[
                                                               {
                                                                   'if': {'column_id': c},
                                                                   'textAlign': 'left'
                                                               } for c in ['job', 'enabled']
                                                           ],
                                                           style_data={
                                                               'whiteSpace': 'normal',
                                                               'color': 'rgb(230, 230, 250)',
                                                               'fontSize': '14px',
                                                               'backgroundColor': 'rgb(112, 128, 144)'
                                                           },
                                                           style_data_conditional=[
                                                               {
                                                                   'if': {'row_index': 'odd'},
                                                                   'backgroundColor': 'rgb(119, 149, 163)',
                                                               },
                                                               {
                                                                   'if': {
                                                                       'filter_query': '{enabled} = False',
                                                                       'column_id': 'enabled'
                                                                   },
                                                                   'backgroundColor': '#eaf044',
                                                                   'color': 'black'
                                                               },
                                                           ],
                                                           style_header={
                                                               'backgroundColor': '#4D4D4D',
                                                               'color': 'rgb(230, 230, 250)',
                                                               'fontWeight': 'bold',
                                                               'fontSize': '16px',
                                                               'border': '1px solid black'
                                                           },
                                                           )
                                  ])
                              ),
                              html.Div(
                                  className="box about",
                                  children=html.Div([
                                       html.B('MARTAS (MagPys automatic real time acquisition system)'),
                                       html.P('Version {}'.format(__version__)),
                                       html.P('written by R. Leonhardt, R. Bailey, R. Mandl, N. Kompein, P. Arneitz, V. Haberle'),
                                       html.P(['MARTAS on GitHUB:',
                                               html.A('click here', href='https://github.com/geomagpy/MARTAS')]),
                                       html.P(['MARTAS manual:',
                                               html.A('click here',
                                                      href='https://github.com/geomagpy/MARTAS?tab=readme-ov-file#martas')]),
                                   ])
                              ),
                              html.Div(
                                  className="box stats",
                                  children=html.Div([
                                     html.B('Configuration'),
                                     html.Div(id='martas-live-update-config'),
                                      dbc.Button(
                                          "Change graph view",
                                          id="horizontal-collapse-button",
                                          className="mb-3",
                                          color="primary",
                                          n_clicks=0,
                                      ),

                                  ])
                             )
                 ]
            )
)

@callback(
    Output("horizontal-collapse1", "is_open"),
     Output("horizontal-collapse2", "is_open"),
    Output("grid-graphs", "style"),
    [Input("horizontal-collapse-button", "n_clicks")],
    [State("horizontal-collapse1", "is_open"),
    State("horizontal-collapse2", "is_open"),],
)
def toggle_collapse(n, is_open, is_open2):
    #print (n)
    if n % 2 == 1:
        sty = {'grid-column' : '1 / 4','grid-row' : '1 / 4'}
    else:
        sty = {'grid-column' : '2 / 4','grid-row' : '2 / 4'}
    if n:
        return not is_open, not is_open2, sty
    return is_open, is_open2, sty


@callback(
    Input('martas-sensors-table', "derived_virtual_data"),
    Input('martas-sensors-table', "derived_virtual_selected_rows"))
def update_graphs(rows, derived_virtual_selected_rows):
    # When the table is first rendered, `derived_virtual_data` and
    # `derived_virtual_selected_rows` will be `None`. This is due to an
    # idiosyncrasy in Dash (unsupplied properties are always None and Dash
    # calls the dependent callbacks when the component is first rendered).
    # So, if `rows` is `None`, then the component was just rendered
    # and its value will be the same as the component's dataframe.
    # Instead of setting `None` in here, you could also set
    # `derived_virtual_data=df.to_rows('dict')` when you initialize
    # the component.
    #print ("------------------------------ UPDATING ROW SELECTORS ")
    if derived_virtual_selected_rows is None:
        derived_virtual_selected_rows = []
    #print (rows, derived_virtual_selected_rows)
    mapa.sensors_to_plot = [rows[i].get('SensorID','') for i in derived_virtual_selected_rows]
    #print (sensors_to_plot)


@callback(Output('martas-live-update-config', 'children'),
              Input('martas-gauge-update', 'n_intervals'))
def update_config(n):
    #print ("updating conf")
    #print ("------------------------------ UPDATING CONFIG ")
    style = {'padding': '5px', 'fontSize': '16px'}
    pos = {}
    htmllist = []
    #print (cfg)
    htmllist.append(html.P(['Station code: {}'.format(mapa.cfg.get('station'))]))
    htmllist.append(html.P(['Sensor configuration: {}'.format(mapa.cfg.get('sensorsconf'))]))
    htmllist.append(html.P(['Buffer directory: {}'.format(mapa.cfg.get('bufferdirectory'))]))
    htmllist.append(html.P(['MQTT broker: {}'.format(mapa.cfg.get('broker'))]))
    return htmllist


@callback([Output('martas-sensors-table', 'data'),
               Output('martas-sensors-table', 'columns')],
              Input('martas-sensors-update', 'n_intervals'))
def martas_update_table(n):
    #print ("updating sensor table")
    #global sensorpath
    #global bufferpath
    #print ("------------------------------ UPDATING TABLE ")
    stable, scols = mapa.get_sensor_table()
    sens_columns = [{'id': c, 'name': c} for c in scols]
    #print ("sens ok")
    return stable, sens_columns


@callback(Output('martas-buffer-gauge', 'value'),
              Input('martas-gauge-update', 'n_intervals'))
def martas_update_gauge_status(n):
    #print ("updating gauge")
    #print ("------------------------------ UPDATING GAUGE ")
    statusdict = mapa.get_storage_usage(debug=False)
    #print ("gauge ok")
    return statusdict['mqtt'].get('used')


@callback([Output('martas-cron-table', 'data'),
               Output('martas-cron-table', 'columns')],
              Input('martas-cron-update', 'n_intervals'))
def martas_update_cron(n):
    #print ("updating cron table")
    #print ("------------------------------ UPDATING CRON ")
    ctable, ccols = mapa.get_cron_jobs()
    cron_columns = [{'id': c, 'name': c} for c in ccols]
    #print ("cron ok")
    return ctable, cron_columns


@callback(
    Output('martas-live-graph', 'figure'),
    Input('martas-graph-update', 'n_intervals'),
    Input('martas-heightslider', 'value'),
    Input('martas-durationslider', 'value'),
    prevent_initial_call=True
)
def martas_update_graph(n, hvalue, duration):
    #print ("ON update", livedata) # no modification, therefore no global definition
    cov = int(duration) * 60
    #print ("Coverage ", cov)
    #print ("------------------------------ PLOTTING DATA ")

    #nd = livedata
    all_names = []
    for el in mapa.livedata:
        if el in mapa.sensors_to_plot:
            sensdict = mapa.livedata.get(el, {})
            if sensdict.get('samplingrate',None):
                shown = sensdict.get('names',[])
                all_names.extend(shown)

    if len(all_names)>0:
        fig = make_subplots(rows=len(all_names), cols=1, vertical_spacing=0.1)
        fig.update_layout(
            plot_bgcolor='#4D4D4D',
            paper_bgcolor='#EDE9E8',
        )
        fig.update_xaxes(
            mirror=True,
            ticks='outside',
            showline=True,
            linecolor='black',
            gridcolor='lightgrey'
        )
        fig.update_yaxes(
            mirror=True,
            ticks='outside',
            showline=True,
            linecolor='black',
            gridcolor='lightgrey'
        )
        fig['layout']['margin'] = {
            'l': 20, 'r': 10, 'b': 10, 't': 10
        }
        fig['layout']['legend'] = {'x': 0, 'y': 1, 'xanchor': 'left'}
        fig.update_layout(height=int(hvalue))

        i = 0
        for f in mapa.livedata:
            if f in mapa.sensors_to_plot:
                ndd = mapa.livedata[f]
                sr = ndd.get('samplingrate',None)
                if sr:
                    amount = int(cov/sr)
                    #print ("Cutting down to", amount, cov, sr)
                else:
                    amount = cov ## assuming 1Hz
                data = ndd.get('data')
                names = ndd.get('names')
                for name in names:
                    i += 1
                    fig.add_trace({
                        'x': data['time'][-amount:],
                        'y': data[name][-amount:],
                        'name': name,
                        'mode': 'lines+markers',
                        'type': 'scatter'
                    }, i, 1)

        return fig

#if __name__ == '__main__':
#    app.run(host="0.0.0.0", debug=False)



#!/usr/bin/env python
"""
a script to communicate with ObsDAQ assuming:
    connected to a PalmAcq
    PalmAcq is in Transparent mode (see manual)
    supported only baud rate 57600
"""
import sys, os, socket, getopt
import serial
import struct, binascii, re, csv
from datetime import datetime, timedelta
from matplotlib.dates import date2num, num2date
import numpy as np
import time

from martas.core import methods as mm


# settings for PalmDaq
PORT = '/dev/ttyUSB0'
BAUDRATE='57600'
eol = '\r'


# settings for ObsDAQ:

#   baudrate 19200 is the default after power on
#   test e.g. by minicom after plugged in

# default clock frequency. There'll be a warning if it is different
    #   98 .. 9.8304MHz
FCLK = '98'

# set 24-bit channel configuration
    # assuming crystal frequency is 9.8304MHz
    # command  $AAnWS0201ccdd 
    # cc ... Range mode (gain)
    #   02 ...  +/-10V
    #   03 ...  +/-5V
    #   04 ...  +/-2.5V
CC = '02'
    # dd ... Data output rate (here examples, see Table 6 and Table 9)
    #   03 .. 3.2 Hz
    #   13 .. 6.4 Hz
    #   23 .. 12.8 Hz
    #   33 .. 19.2 Hz
    #   43 .. 32.0 Hz
    #   53 .. 38.4 Hz
    #   63 .. 64 Hz
    #   72 .. 76.8 Hz
    #   82 .. 128 Hz
    #   92 .. 640 Hz
    #   A1 .. 1280 Hz

DD = '23'
#DD = '63'

# setting internal trigger timing
    # command $AAPPeeeeffff
    # eeee ... triggering interval
    # ffff ... low-level time
    # recommended by Mingeo (9.8304MHz quartz crystal, see Table 9)
    #   DD EEEE FFFF Samples/s FilterFrequency
    #   63 0BFF 026D 50Hz      64Hz 
    #   72 09FF 01FD 60Hz      76.8Hz 
    #   82 05FF 011D 100Hz     128Hz 
    #   92 04AF 038D 128Hz     640Hz 
    #   92 03FF 02DD 150Hz     640Hz 
    #   92 02FF 01DD 200Hz     640Hz
EEEE = '3C00'
FFFF = '0600'
#EEEE = '0BFF'
#FFFF = '026D'

# program offset calibration constants
    # command $AAnWOaaaaaa
#OFFSET = ['','','']
OFFSET = ['FFF19A','FFF41B','FFF70C']

# program full-scale calibration constants
    # command $AAnWFffffff
#FULLSCALE = ['','','']
FULLSCALE = ['3231C0','32374B','323A7E']

# ----------------------------------
# please don't edit beyond this line
# ----------------------------------

global ser
global QUIET
QUIET = False

def lineread(ser,eol):
            # FUNCTION 'LINEREAD'
            # Does the same as readline() plus timeout
            ser_str = ''
            timeout = time.time()+2
            if sys.version_info >= (3, 0):
                eol=eol.encode('ascii')
                while True:
                    byte = ser.read()
                    if byte == eol:
                        break
                    if time.time() > timeout:
                        print ('Timeout')
                        break
                    try:
                        ser_str += byte.decode('ascii')
                    except:
                        print('obsdaq.py: decode error, got '+str(byte))

            else:
                while True:
                    char = ser.read()
                    if char == eol:
                        break
                    if time.time() > timeout:
                        print ('Timeout')
                        break
                    ser_str += char
            return ser_str

def send_command(ser,command,eol,hex=False):
    #command = eol+command+eol
    command = command+eol
    #print 'Command:  %s \n ' % command.replace(eol,'')
    sendtime = date2num(datetime.utcnow())
    #print "Sending"
    if sys.version_info >= (3, 0):
        ser.write(command.encode('ascii'))
    else:
        ser.write(command)
    #print "Received something - interpretation"
    response = lineread(ser,eol)
    #print "interprete"
    receivetime = date2num(datetime.utcnow())
    meantime = np.mean([receivetime,sendtime])
    #print "Timediff", (receivetime-sendtime)*3600*24
    return response, num2date(meantime).replace(tzinfo=None)


def command(call):
    global ser
    global QUIET
    if not QUIET:
        print(call)
    answer, actime = send_command(ser,call,eol)
    if not QUIET:
        print(answer)
    return answer


def main(argv):
    try:
        opts, args = getopt.getopt(argv,"hvm:cn:f:apsodiq",[])
    except getopt.GetoptError:
        print ('unknown option')
        sys.exit(2)

    global ser
    port = PORT
    baudrate = BAUDRATE
    cc = CC
    dd = DD
    eeee = EEEE
    ffff = FFFF
    offset = OFFSET
    fullscale = FULLSCALE
    for opt, arg in opts:
        if opt == '-h':
            print ('-------------------------------------')
            print ('Description:')
            print ('Sending data to ObsDAQ via PalmAcq')
            print ('  therefore it is necessary to bring PalmAcq in Transparent mode')
            print ('  i.e. by using the palmacq.py script:')
            print ('  python palmacq.py -t')
            print ('-------------------------------------')
            print ('Usage:')
            print ('obsdaq.py -q -v -m config-file -c -n channel -f channel -a -p -s -o -d -i')
            print ('-------------------------------------')
            print ('Options:')
            print ('-v          : show version of ObsDAQ and quit')
            print ('-m          : MARTAS compatible config file')
            print ('')
            print ('-c          : define calibration constants')
            print ('-n channel  : perform an offset system calibration (input must be 0) of channel 1-3')
            print ('-f channel  : perform a full-scale calibration (input must be maximum) of channel 1-3')
            print ('')
            print ('-a          : start acquisition (calibrate first!)')
            print ("-p          : exit free run or triggered mode - stop acquisition")
            print ('-s          : show output from serial line')
            print ('-o          : show the formatted output')
            print ('')
            print ('-d          : show definitions made in the code resp. config file')
            print ('-i          : show info about ObsDAQ settings')
            print ("-q          : quiet: don't show commands and answers. Has to be first option.")
            print ('-------------------------------------')
            print ('Examples:')
            print ('python obsdaq.py -t')
            sys.exit()
        if opt in ("-m"):
            configfile = os.path.abspath(arg)
            conf = mm.get_conf(configfile)
            port = conf.get('port')
            baudrate = conf.get('baudrate')
            cc = str(conf.get('CC')).zfill(2)
            dd = str(conf.get('DD')).zfill(2)
            eeee = str(conf.get('EEEE')).zfill(4)
            ffff = str(conf.get('FFFF')).zfill(4)
            offset = [str(conf.get('OFFSETX')),str(conf.get('OFFSETY')),str(conf.get('OFFSETZ'))]
            fullscale = [str(conf.get('FULLSCALEX')),str(conf.get('FULLSCALEY')),str(conf.get('FULLSCALEZ'))]

        ser = serial.Serial(port, baudrate=baudrate , parity='N', bytesize=8, stopbits=1, timeout=2)

        # some calculations
        FCLKdic = {'98':9.8304e6,'92':9.216e6,'76':7.68e6}
        fclkFreq = FCLKdic[FCLK]
        # factor for trigger timing parameters eeee and ffff
        micros = 64./fclkFreq
        CCdic = {'02':10.,'03':5.,'04':2.5}
        gain = CCdic[cc]
        DDdic = {'03':3.2,'13':6.4,'23':12.8,'33':19.2,'43':32.,'53':38.4,'63':64.,'72':76.8,'82':128.,'92':640.,'A1':1280.}
        # data output rate
        drate = DDdic[dd] * FCLKdic[FCLK] / FCLKdic['98']
        eeeeTime = int('0x'+eeee,16)*micros
        ffffTime = int('0x'+ffff,16)*micros

        if opt in ("-v", "--version"):
            answer = ser.read(10)
            if answer:
                print ('ObsDaq: it seems, that acquisition is in progress.')
                print ('ObsDaq: please stop acquisition using -p or --stop option first!')
                exit(2)
            print ('ObsDaq: Firmware version and firmware date')
            command('$01F')
            print ('ObsDaq: Module name (firmware version)')
            command('$01M')
        if opt in ("-q", "--quiet"):
            global QUIET
            QUIET = True
        if opt in ("-p", "--stop"):
            # stop acquisition
            print('ObsDaq: trying to stop acquisition...')
            command('#01ST')
            stopped = False
            while not stopped:
                # quick stop
                answer, actime = send_command(ser,'||||||||||||||||||||||','')
                if answer:
                    print ('Answer from ObsDAQ:')
                    print (answer)
                    print ('ObsDaq: please wait!')
                    for i in range(100):
                        ser.write('|||||||||||||||||||||||||||||||||||\r')
                else:
                    print ('ObsDaq: stopped')
                    stopped = True    
            # stop free run or triggered mode, enter idle mode
            command('#01ST')
        elif opt in ("-i", "--info"):
            answer = ser.read(10)
            if answer:
                print ('ObsDaq: it seems, that acquisition is in progress.')
                print ('ObsDaq: please stop acquisition using -p or --stop option first!')
                exit(2)
            print ('Information - please refer to ObsDAQ manual')
            print ('Serial number')
            sn = command('$01SN')
            sn = sn.split('SN')[1]
            print ('S/N: '+sn)
            # turn off triggering to enable getting 24 bit configuration
            print ('turning off triggering')
            command('#01PP00000000')
            # wait a second
            time.sleep(1)
            print ('baud rate code')
            command('$012')
            print ('quarz frequency')
            ans = command('$01QF')
            ans = ans.split('R')[1]
            if not ans == FCLK:
                print ('WARNING: this code is not written for this crystal frequency.')
                print ('   please adopt python code!')
                print ('')
            print ('first EEPROM command')
            print ('  *ERR indicates there is no EEPROM programmed for autostart')
            command('$01IR0')
            print ('24bit channel 0 config') 
            command('$010RS')
            print ('24bit channel 1 config')
            command('$011RS')
            print ('24bit channel 2 config') 
            command('$012RS')
            
            # get 24-bit channel configuration
            print ('channel configuration')
            command('$010RS')
            command('$011RS')
            command('$012RS')
            time.sleep(1)

            # get offset calibration constants
            print ('offset calibration constants')
            command('$010RO')
            command('$011RO')
            command('$012RO')

            # get full-scale calibration constants
            print ('full scale calibration constants')
            command('$010RF')
            command('$011RF')
            command('$012RF')

        elif opt in ("-s", "--show"):
            while True:
                print (lineread(ser,eol))

        elif opt in ("-c", "--calib"):
            answer = ser.read(10)
            if answer:
                print ('ObsDaq: it seems, that acquisition is in progress.')
                print ('ObsDaq: please stop acquisition using -p or --stop option first!')
                exit(2)

            print ('ObsDaq: turning off triggering')
            command('#01PP00000000')
            # wait a second
            time.sleep(1)

            print ('ObsDaq: setting 24-bit channel configuration')
            # cc=02..+/-10V
            # dd=23..12.8Hz
            command('$010WS0201'+cc+dd)
            time.sleep(1)
            command('$011WS0201'+cc+dd)
            time.sleep(1)
            command('$012WS0201'+cc+dd)
            time.sleep(1)

            # calibration constants
            for i in range(3):
                if offset[i]:
                    print ('programming given offset calibration constant for channel '+str(i+1))
                    command('$01'+str(i)+'WO'+offset[i])
            for i in range(3):
                if fullscale[i]:
                    print ('programming given full-scale calibration constant for channel '+str(i+1))
                    command('$01'+str(i)+'WF'+fullscale[i])
            time.sleep(1)
            # execute an offset and full-scale self-calibration if necessary
            for i in range(3):
                if not offset[i] or not fullscale[i]:
                    print ('executing an offset and full-scale self-calibration for channel '+str(i+1))
                    command('$01'+str(i)+'WCF0')
                    print ('calibrating...')
                    # check calibration finished?
                    calfin = False
                    while not calfin:
                        time.sleep(0.5)
                        ans = command('$01'+str(i)+'RR')
                        ans = ans.split('R')[1]
                        if ans == '0':
                            print ('ObsDaq: finished calibration')
                            calfin = True
            
            # get 24-bit channel configuration
            print ('channel configuration')
            a=command('$010RS')
            b=command('$011RS')
            c=command('$012RS')
            if QUIET:
                print('\t'+a+'\t'+b+'\t'+c)
                accdd='\t+/-'+str(CCdic[a[7:9]])+'V  '+str(DDdic[a[9:11]])+'Hz'
                bccdd='\t+/-'+str(CCdic[b[7:9]])+'V  '+str(DDdic[b[9:11]])+'Hz'
                cccdd='\t+/-'+str(CCdic[c[7:9]])+'V  '+str(DDdic[c[9:11]])+'Hz'
                print(accdd+'\t'+bccdd+'\t'+cccdd)
            time.sleep(1)

            # get offset calibration constants
            print ('offset calibration constants')
            a=command('$010RO')
            b=command('$011RO')
            c=command('$012RO')
            if QUIET:
                print('\t'+a+'\t'+b+'\t'+c)

            # get full-scale calibration constants
            print ('full scale calibration constants')
            a=command('$010RF')
            b=command('$011RF')
            c=command('$012RF')
            if QUIET:
                print('\t'+a+'\t'+b+'\t'+c)

        elif opt in ("-n", "--offsetcal"):
            try:
                ch = int(arg)-1
                ch = str(ch)
            except:
                print ("channel must be in 1..3")
                exit(2)
            answer = ser.read(10)
            if answer:
                print ('it seems, that acquisition is in progress.')
                print ('please stop acquisition using -p or --stop option first!')
                exit(2)
            
            print ('turning off triggering')
            command('#01PP00000000')
            # wait a second
            time.sleep(1)

            print ('setting 24-bit channel configuration')
            # cc=02..+/-10V
            # dd=23..12.8Hz
            command('$01'+ch+'WS0201'+cc+dd)
            time.sleep(1)

            print ('performing offset calibration of channel '+arg)
            command('$01'+ch+'WCF3')
            time.sleep(1)
            print ('get the constants from python obsdac.py -i')

        elif opt in ("-f", "--fullscalecal"):
            try:
                ch = int(arg)-1
                ch = str(ch)
            except:
                print ("channel must be in 1..3")
                exit(2)
            answer = ser.read(10)
            if answer:
                print ('it seems, that acquisition is in progress.')
                print ('please stop acquisition using -p or --stop option first!')
                exit(2)
            
            print ('turning off triggering')
            command('#01PP00000000')
            # wait a second
            time.sleep(1)

            print ('setting 24-bit channel configuration')
            # cc=02..+/-10V
            # dd=23..12.8Hz
            command('$010WS0201'+cc+dd)
            time.sleep(1)
            command('$011WS0201'+cc+dd)
            time.sleep(1)
            command('$012WS0201'+cc+dd)
            time.sleep(1)

            print ('performing full-scale calibration of channel '+arg)
            command('$01'+ch+'WCF4')
            time.sleep(1)
            print ('get the constants from python obsdac.py -i')



        elif opt in ("-d", "--defs"):
            # show user settings human readable
            print ("Definitions for ObsDAQ made in this file:")
            print ('')
            print ('Clock frequency:')
            print ('FCLK:\t'+FCLK)
            print ('\t'+str(fclkFreq/1000000.)+'MHz')
            print ('Gain:')
            print ('CC:\t'+cc)
            print ('\t+/-'+str(gain)+'V')
            print ('Data output rate:')
            print ('DD:\t'+dd)
            print ('\t'+str(1./drate)+" s")
            print ('\t'+str(drate)+' Hz')
            print ('triggering interval')
            print ('EEEE:\t'+eeee)
            print ('\t'+str(eeeeTime)+" s")
            print ('\t'+str(1./eeeeTime)+" Hz")
            print ('low-level time')
            print ('FFFF:\t'+ffff)
            print ('\t'+str(ffffTime)+" s")
            print ('\t'+str(1./ffffTime)+" Hz")
            # there are limitations in choosing DD, EEEE and FFFF:
            if drate < 1./eeeeTime:
                print ('WARNING: Digital filter output rate is smaller than the triggering frequency!')
                print ('drate '+str(drate)+'Hz < 1./eeeeTime '+str(1./eeeeTime)+'Hz')
            if 1./drate - 0.0011 > eeeeTime:
                print ('WARNING: The difference of triggering interval and digital filter interval is smaller than 1.1ms!')
                print ('1./drate '+str(1./drate)+'s - 0.0011s > eeeeTime '+str(eeeeTime)+'s')
            if eeeeTime - ffffTime < 0.00019:
                print ('WARNING: Difference between triggering interval and low-level time is '+str(eeeeTime-ffffTime)+'!')
            if eeeeTime - ffffTime < 1./drate:
                print ('WARNING: Triggering interval - Low-level time is higher than the Digital filter output rate!')

        elif opt in ("-a", "--start"):
            print ('ObsDaq: starting acquisition')
            # set internal trigger timing (Table 9)
            command('#01PP'+eeee+ffff)
            # wait a second
            time.sleep(2)

            # start acquisition (plus supplementary data every second)
            command('#01CS')
            
            time.sleep(2)



if __name__ == "__main__":
    main(sys.argv[1:])
            






    



#!/usr/bin/python

"""
Simple method to initialize GSM sensors (GSM19, GSM90)
"""

import serial
from datetime import datetime, timezone
import binascii
import sys, getopt

import time


def lineread(ser,eol=None,debug=False):
    """
    DESCRIPTION:
       Does the same as readline(), but does not require a standard 
       linebreak character ('\r' in hex) to know when a line ends.
       Variable 'eol' determines the end-of-line char: '\x00'
       for the POS-1 magnetometer, '\r' for the envir. sensor.
       (Note: required for POS-1 because readline() cannot detect
       a linebreak and reads a never-ending line.)
    PARAMETERS:
       eol:   (string) lineend character(s): can be any kind of lineend
                           if not provided, then standard eol's are used
    """
    if sys.version_info >= (3,0):
        eol = eol.decode()
    
    if not eol:
        eollist = ['\r','\x00','\n']
    else:
        eollist = [eol]

    ser_str = ''
    while True:
        char = ser.read()
        if sys.version_info >= (3,0):
            char = char.decode()
        if debug:
            print ("receiving:", char)
        if char in eollist:
            break
        ser_str += char
    return ser_str


def hexify_command(command,eol):
    """
    DESCRIPTION:
       This function translates the command text string into a hex
       string that the serial device can read. 'eol' is the 
       end-of-line character. '\r' for the environmental sensor,
       '\x00' for the POS-1 magnetometer.
    This a  python2 only method    
    """
    commandstr = []
    for character in command:
            hexch = binascii.hexlify(character)
            commandstr.append(('\\x' + hexch).decode('string_escape'))
    command_hex = ''.join(commandstr) + (eol)
    return command_hex


def send_command(ser,command,eol=None,hexify=False,bits=0,debug=False):
    """
    DESCRIPTION:
        General method to send commands to a e.g. serial port.
        Returns the response
    PARAMETER:
        bits: (int) provide the amount of bits to be read
        eol:  (string) end-of-line string
    Options:
        POS1:           hexify=True, line=True, eol= '\x00'
        ENV05:          hexify=True, line=True, eol='\r'
        GSM90Sv6:       hexify=False, line=False, eol=''
        GSM90Sv7:       hexify=False, line=False, eol
        GSM90Fv7:       hexify=False, line=False, eol
    """
    print('-- Sending command:  ', command, eol)
    if sys.version_info >= (3,0):
        command = command.encode('ascii')
        if eol:
            eol = eol.encode('ascii')
    if sys.version_info >= (3,0) and hexify:
        # python3, hexify selected
        if debug:
            print ("Hexify with py3 selected")
        ser.write(command)
        if eol:
            ser.write(eol)
    elif hexify:
        # python2 with hexify
        if debug:
            print ("Hexify with py2 selected")
        command = hexify_command(command,eol)
        ser.write(command)
    else:
        if debug:
            print ("Default write selected")
        if not eol:
            ser.write(command)
        else:
            ser.write(command+eol)
    if bits==0:
        response = lineread(ser,eol,debug=debug)
    else:
        response = ser.read(bits)
        if sys.version_info >= (3,0):
            response = response.decode()
    print('-- Response: ', response)
    return response


def main(argv):
    # Communication parameter defaults
    baudrate = 115200
    port = '/dev/ttyS0'
    parity='N'
    bytesize=8
    stopbits=1
    timeout=5
    #flowcontrol=  ## Not used so far

    # General defaults (time, etc)
    command = ''
    commands = []
    responseaction = []
    hexify = False
    eol = None
    bits = 0
    timeformat = None
    hwflwctrl = False
    debug = False

    try:
        opts, args = getopt.getopt(argv,"hb:p:f:y:a:s:o:c:r:xe:d:i:k:D",["baudrate=","port=","flowcontrol=","bytesize=","parity=","stopbits=","timeout=","command=","responseaction=", "eol=", "eol-ord=", "bits2read=", "timeformat=",])
    except getopt.GetoptError:
        print('Check your options:')
        print('serial-init.py -b <baudrate> -p <port> -f <flowcontrol> -y <bytesize> -a <parity> -s <stopbits> -o <timeout> -c <commands> -r <responseaction> -x <hexify> -e <eol> -d <eol-ord> -i <bits2read> -k <timeformat>')
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print('------------------------------------------------------')
            print('Usage:')
            print('serial-init.py -b <baudrate> -p <port> -f <flowcontrol> -y <bytesize> -a <parity> -s <stopbits> -o <timeout> -c <commands> -r <responseaction> -x <hexify> -e <eol> -d <eol-ord> -i <bits2read> -k <timeformat>')
            print('------------------------------------------------------')
            print('Options:')
            print('-h                             help')
            print('-b                             set baudrate - default is 115200')
            print('-p                             set port - default is /dev/ttyS0')
            print('-f                             set flowcontrol')
            print('-y  e.g. SEVENBITS             set bytesize - default is 8     ') 
            print('-a  e.g. PARITY_EVEN           set parity   - default is N     ')
            print('-s                             set stopbits - default is 1     ')
            print('-o                             set timeout  - default is 5 (sec) ')
            print('-c  e.g.C,1611104142567,T48    comma separated string with commands')
            print('-r  e.g. z-save,z              comma separated list of two items: ')
            print('                               if the first item is found within ')
            print('                               the response of a command, then the ')
            print('                               second item is send')
            print('-x                             hexify commands                 ')
            print('-e                             end-of-line characters to be used')
            print('-d                             end-of-line character as ascii number')
            print('                               defaults are: ')
            print('-i                             define the bits size of the response')
            print('                               default is to read until eol (option e)')
            print('-k  e.g. "%y%m%d%w%H%M%S"      set timestamp-format, if provided and placeholder')
            print('                               "datetime" is found in commands, then it is replaced')
            print('                               by a current UTC timestamp of the provided format')
            print('                               e.g. -c C,datetime,T48 -> -c C,1611104153425,T48')
            
            print('------------------------------------------------------')
            print('Examples:')
            print('1. Initializing POS1:')
            print('   python serialinit.py -p "/dev/ttyS1" -b 9600 -c "mode text,time datetime,date 11-22-16,range 48500,auto 5" -k "%H:%M:%S" -x -d 0 -f True')
            print('2. Initializing GSM90Sv6:')
            print('   python serialinit.py -p "/dev/ttyUSB0" -c S,5,T048,C,datetime,R -k "%y%m%d%w%H%M%S" -i 1024')
            print('3. Initializing GSM90Sv7:')
            print('   python serialinit.py -p "/dev/ttyUSB0" -c S,5,T048.5,C,datetime,D,R -k "%y%m%d%w%H%M%S" -r "z-save,z" -i 1024') 
            print('4. Initializing BM35:')
            print('   python serialinit.py -b 2400 -p "/dev/ttyS1" -c "A00d03000" -d 13')
            sys.exit()
        elif opt in ("-b", "--baudrate"):
            baudrate = arg
        elif opt in ("-p", "--port"):
            port = arg
        elif opt in ("-y", "--bytesize"):
            try:
                bytesize = int(arg)
            except:
                bytesize = arg
        elif opt in ("-a", "--parity"):
            parity = arg
        elif opt in ("-s", "--stopbits"):
            try:
                stopbits = int(arg)
            except:
                stopbits = arg
        elif opt in ("-o", "--timeout"):
            try:
                timeout = int(arg)
            except:
                timeout = arg
        elif opt in ("-e", "--eol"):
            eol = arg
        elif opt in ("-d", "--eol-ord"):
            eol = chr(int(arg))
        elif opt in ("-c", "--commands"):
            command = arg
        elif opt in ("-r", "--responseaction"):
            respac = arg
            responseaction = respac.split(',')
            if not len(responseaction) == 2:
                print ("serial-init: warning... cannot interpret (r) responseaction - skipping this option")
                responseaction = []
        elif opt in ("-x", "--hexify"):
            hexify = True
        elif opt in ("-i", "--bits2read"):
            try:
                bits = int(arg)
            except:
                bits = 0
                print ("serial-init: warning... integer number expected for (i) - skipping this option")
        elif opt in ("-k", "--timeformat"):
            timeformat = arg
        elif( opt in ("-f", "--flowcontrol")):
            hwflwctrl = True
        elif opt in ("-D", "--debug"):
            debug = True

    # Re-format command sequence
    if command.find('datetime') > 0:
        if not timeformat:
            print ("serial-init: warning... found 'datetime' placeholder in command sequence, but no timeformat is provided (-k):")
            print ("             ... skipping option")

        else:
            try:
                currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
                timestamp = str(datetime.strftime(currenttime,timeformat))
                command = command.replace('datetime',timestamp)
            except:
                print ("serial-init: warning... could not interpret datetimeformat provided by (-k) ... skipping this")
                pass
    commands = command.split(',')

    # Connecting
    print ("Welcome to serial-init")
    print ("------------------------------------------")
    print ("Establishing connection to serial port:")
    try:
        print ("TEST", baudrate,parity,bytesize,stopbits,timeout)
        ser = serial.Serial(port, baudrate=baudrate, parity=parity, bytesize=bytesize, stopbits=stopbits, timeout=timeout)

        ########
        # checking if output or input buffer contains waiting bytes
        # if yes reset output buffer
        ########
        if( not ( ser.out_waiting == 0)):
            print( 'Flushing output buffer for initialization...')
            ser.flushOutput()
            print( '...done')
        if( not ( ser.inWaiting() == 0)):
            print( 'Flushing input buffer for initialization...')
            ser.flushInput()
            print( '...done flushing')
        # activating serial connection hardware flow-control
        if( hwflwctrl):
            ser.rts = True
            ser.dtr = True
            print( 'Hardware flowcontrol activated by ser.rts and ser.dtr...')
        ######
        print('.... Connection made.')
    except Exception as ex: # changed output string to respond with Exception
        print('.... Connection flopped with Exception:{}.'.format( ex))
        print('--- Aborting ---')
        sys.exit(2)
    print('')

    # Sending/Receiving
    print ("------------------------------------------")
    if len(commands) > 0:
        print ("Sending command sequence:")
        for item in commands:
            response = send_command(ser, item, eol=eol, hexify=hexify, bits=bits,debug=debug)
            if len(responseaction) == 2:
                if response.find(responseaction[0]) > 0:
                    print ("TEST:", responseaction)
                    print ("... found matching response - performing response-action")
                    response = send_command(ser, responseaction[1], eol=eol, hexify=hexify, bits=bits,debug=debug)
        print('')
        print ("... Done ... good bye")
    else:
        print ("No commands provided:")
        amount = 1024
        print ("... reading {} bits to check whether data is send on port {}:".format(amount,port))
        response = ser.read(amount)
        print (response)
        print('')
        print ("... Done ... good bye")
        sys.exit(2)

if __name__ == "__main__":
   main(sys.argv[1:])




#!/usr/bin/env python
# coding=utf-8

"""
MARTAS - threshold application
################################

DESCRIPTION:
Threshold application reads data from a defined source
(DB or file, eventually MQTT).
Threshold values can be defined for keys in the data file,
notifications can be send out when certain criteria are met,
and switching commands can be send if thresholds are exceeded
or undergone.
All threshold processes can be logged and  can be monitored
by nagios, icinga or martas.
Threshold.py can be scheduled in crontab.

REQUIREMENTS:# Proxies (if deviating from system proxies)
#https         :   http://192.168.1.1:3128

pip install geomagpy (>= 2.0.0)


APPLICATION:
threshold -m /path/to/threshold.cfg


switch.cfg: (looks like)
##  ----------------------------------------------------------------
##           CONFIGURATION DATA for THRESHOLD.PY
##  ----------------------------------------------------------------

# MARTAS directory
martasdir            :   /home/cobs/MARTAS/

# Define data source (file, db, ...)
source               :   file

# If source = db then define data base credentials created by addcred (MARTAS)
dbcredentials        :   None


# If source = file define the MARTAS buffer base path
bufferpath           :   /srv/mqtt/


# Logfile (a json style dictionary, which contains statusmessages)
#logfile              :   /var/log/magpy/threshold.log
logfile              :   /home/leon/Tmp/threshold.log


# Notifaction (uses martaslog class, one of email, telegram, mqtt, log)
notification         :   email
notificationconfig   :   /etc/martas/notification.cfg


# Report level ("full" will report all changes, also within range states)
reportlevel          :   partial


# serial communication for switch commands (making use of ardcomm.py (MARTAS/app)
#port                :   device to which command is send - default is /dev/ttyACM0
#baudrate            :   default is 9600
#parity              :   default is "N"
#bytesize            :   default is 8
#stopbits            :   default is 1
#timeout             :   default is 2
#eol                 :   end of line - default is \r\n


#parameter (all given parameters are checked in the given order, use semicolons for parameter list):
# sensorid; timerange to check; key to check, value, function, state, statusmessage, switchcommand(optional)
1  :  DS18B20XX;1800;t1;5;average;below;default
2  :  DS18B20XX;1800;t1;5;average;below;none
3  :  DS18B20XZ;600;t2;10;max;below;ok
4  :  DS18B20XZ;600;t2;10;max;above;warning at week
5  :  DS18B20XZ;600;t2;20;max;above;alarm issued at date
6  :  DS18B20XZ;600;t2;3;stddev;above;flapping state
7  :  DS18B20XX;1800;t1;10;median;above;default;swP:1:4
10  :  DS18B20XX;1800;t1;10;median;below;default;swP:0:4

#to be continued...

# SensorID, key:  if sensorid and key of several lines are identical, always the last valid test line defines the message
#                 Therefore use warning thresholds before alert thresholds
# Function:       can be one of max, min, median, average(mean), stddev
# State:          can be one below, above, equal
# Statusmessage:  default is replaced by "Current 'function' 'state' 'value', e.g. (1) "Current average below 5"
#                 the following words (last occurrence) are replace by datetime.utcnow(): date, month, year, (week), hour, minute
#                 "date" is replaced by current date e.g. 2019-11-22
#                 "month" is replaced by current month e.g. 2019-11
#                 "week" is replaced by current calender week e.g. 56
#                 "minute" looks like 2019-11-22 13:10
#                 -> "date" changes the statusmessage every day and thus a daily notification is triggered as long a alarm condition is active


## Description: Parameterset "1":  if temperature t1 (average of last 30 min (1800 sec)) is falling below 5 degrees
## then send default statusmessage to the notification system (e.g. email)
## and eventually send switchcommand to serial port
## IMPORTANT: statusmessage should not contain semicolons, colons and commas; generally avoid special characters


You can also execute basic shell scripts dependening on thresholds:
10  :  DS18B20XX;1800;var3;10;median;below;default;;/home/user/shutdown.sh

with a shutdown.sh like:
#!/bin/bash
DATE=$(date +%Y%m%d%H%M)
echo "${DATE}: Shutting down in 1 minute" >> /var/log/magpy/shutdown.log
# shutdown in one minute
shutdown -P +1


"""

# Define packges to be used (local refers to test environment)
# ------------------------------------------------------------
from magpy.stream import DataStream, read
from magpy.core import database
from magpy.core.methods import is_number
import magpy.opt.cred as mpcred
from martas.core import methods as mameth
from martas.core.methods import martaslog as ml

from datetime import datetime, timedelta, timezone
import sys, getopt, os
import paho.mqtt.client as mqtt
import socket

pyvers = '3'


class sp(object):
    # Structure for switch parameter
    configdict = {}
    configdict['bufferpath'] = '/srv/mqtt'
    configdict['dbcredentials'] = 'cobsdb'
    configdict['database'] = 'cobsdb'
    configdict['reportlevel'] = 'partial'
    configdict['notification'] = 'log'
    valuenamelist = ['sensorid','timerange','key','value','function','state','statusmessage','switchcommand','executecommand']
    #valuedict = {'sensorid':'DS18B20','timerange':1800,'key':'t1','value':5,'function':'average','state':'below','message':'on','switchcommand':'None'}
    #parameterdict = {'1':valuedict}
    parameterdict = {}
    configdict['version'] = '1.0.0' # thresholdversion
    configdict['martasdir'] = '/home/cobs/MARTAS/'
    # Serial Comm
    configdict['port'] = '/dev/ttyACM0'
    configdict['baudrate'] = 9600
    configdict['parity'] = 'N'
    configdict['bytesize'] = 8
    configdict['stopbits'] = 1
    configdict['timeout'] = 2
    configdict['eol'] = '"\r\n"'

    #Testset
    #configdict['startdate'] = datetime(2018,12,6,13)
    #valuenamelist = ['sensorid','timerange','key','value','function','state','statusmessage','switchcommand']
    #valuedict = {'sensorid':'ENV05_2_0001','timerange':1800,'key':'t1','value':5,'function':'average','state':'below','message':'on','switchcommand':'None'}


def get_data(source, path, dbcredentials, sensorid, amount, startdate=None, debug=False):
    """
    DESCRIPTION:
    read the appropriate amount of data from the data file, database or mqtt stream
    """

    data = DataStream()
    msg = ''
    if not startdate:
        startdate = datetime.now(timezone.utc).replace(tzinfo=None)
        endtime = None
    else:
        endtime = startdate

    starttime = startdate-timedelta(seconds=int(amount))

    if source in ['file','File']:
        # TODO eventually check for existance and look for similar names
        # expath = CheckPath(filepath)
        # if expath:
        if debug:
            print ("Trying to access files {} in {}: Timerange: {} to {}".format(sensorid, path,starttime,endtime))
        try:
            data = read(os.path.join(path, sensorid, '*'), starttime=starttime, endtime=endtime)
        except:
            msg = "Could not access data for sensorid {}".format(sensorid)
            if debug:
                print (msg)
    elif source in ['db','DB','database','Database']:
        db = mameth.connect_db(dbcredentials)
        data = db.read(sensorid, starttime=starttime)

    if debug:
        print ("Got {} datapoints".format(data.length()[0]))

    return (data, msg)

def get_test_value(data=DataStream(), key='x', function='average', debug=False):
    """
    DESCRIPTION
    Returns comparison value(e.g. mean, max etc)
    """
    if debug:
        print ("Obtaining test value for key {} with function {}".format(key,function))
    func = 'mean'
    testvalue = None
    msg = ''
    n = data.length()[0]
    keys = data._get_key_headers()

    if not key in keys:
        print ("Requested key not found")
        return (testvalue, 'failure')
    if function in ['mean','Mean','average', 'Average','Median','median']:
        if n < 3:
            print ("not enough data points --- {} insignificant".format(function))
        if function in ['mean','Mean','average', 'Average']:
            func = 'mean'
        elif function in ['Median','median']:
            func = 'median'
        testvalue = data.mean(key,meanfunction=func)
    elif function in ['max','Max']:
        testvalue = data._get_max(key)
    elif function in ['min','Min']:
        testvalue = data._get_min(key)
    elif function in ['stddev','Stddev']:
        mean,testvalue = data.mean(key,std=True)
    else:
        msg = 'selected test function not available'

    if debug:
        print (" ... got {}".format(testvalue))

    return (testvalue, msg)


def check_threshold(testvalue, threshold, state, debug=False):
    """
    DESCRIPTION:
     returns statusmessage
    """
    evaluate = False
    msg = ''
    comp = "=="

    if state in ['below','Below','smaller']:
        comp = '<'
    elif state in ['above','Above','greater']:
        comp = '>'
    elif state in ['equal','Equal']:
        comp = '=='
    elif state in ['equalabove']:
        comp = '>='
    elif state in ['equalbelow']:
        comp = '<='
    else:
        msg = 'state needs to be one of below, above or equal'

    tester = "{} {} {}".format(testvalue, comp, threshold)
    if debug:
        print ("Checking: {}".format(tester))
    try:
        if eval(tester):
            evaluate = True
            if debug:
                print (" ... positive")
        else:
            evaluate = False
            if debug:
                print (" ... negative - criteria not met")
    except:
        msg = "Comparison {} failed".format(tester)

    return (evaluate, msg)

def interprete_status(valuedict, debug=False):
    """
    DESCRIPTION:
    checks the message and replace certain keywords with predefined test
    """
    msg = valuedict.get('statusmessage')
    defaultline =  'Current {} {} {}'.format(valuedict.get('function'),valuedict.get('state'),valuedict.get('value'))
    ct = datetime.now(timezone.utc).replace(tzinfo=None)
    msg = msg.replace('none', '')
    msg = msg.replace('default', defaultline)
    msg = msg.replace('date', datetime.strftime(ct,"%Y-%m-%d"))
    msg = msg.replace('hour', datetime.strftime(ct,"%Y-%m-%d %H:%M"))
    msg = msg.replace('minute', datetime.strftime(ct,"%Y-%m-%d %H:%M"))
    msg = msg.replace('month', datetime.strftime(ct,"%Y-%m"))
    msg = msg.replace('year', datetime.strftime(ct,"%Y"))
    msg = msg.replace('week', 'week {}'.format(ct.isocalendar()[1]))

    return msg


def assign_parameterlist(namelist=None,confdict=None):
    """
    DESCRIPTION:
        assign the threshold parameters to a dictionary
    EXAMPLE:
        para = assign_parameterlist(sp.valuenamelist,conf)
    """
    if not namelist:
        namelist = []
    if not confdict:
        confdict = {}
    para = {}
    for i in range(1,9999):
        valuedict = {}
        valuelist = confdict.get(str(i),[])
        if len(valuelist) > 0:
            if not len(valuelist) in [len(namelist), len(namelist)-1, len(namelist)-2]:
                # check whether all values (except the two optional ones are present)
                print ("PARAMETER: provided values differ from the expected amount - please check")
            else:
                for idx,val in enumerate(valuelist):
                    valuedict[namelist[idx]] = val.strip()
                para[str(i)] = valuedict
    return para

def main(argv):

    para = sp.parameterdict
    conf = sp.configdict
    debug = False
    configfile = None
    statusdict = {}
    statuskeylist = []
    travistestrun = False
    changes = []
    proxies = {}
    hostname = socket.gethostname().upper()


    usagestring = 'threshold.py -h <help> -m <configpath>'
    try:
        opts, args = getopt.getopt(argv,"hm:DT",["configpath="])
    except getopt.GetoptError:
        print ('Check your options:')
        print (usagestring)
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print ('------------------------------------------------------')
            print ('Usage:')
            print (usagestring)
            print ('------------------------------------------------------')
            print ('Options:')
            print ('-h            help')
            print ('-m            Define the path for the configuration file.')
            print ('              Please note: a configuration file is obligatory')
            print ('              ----------------------------')
            print ('              configurationfile')
            print ('              ----------------------------')
            print ('              threhold.cfg: (looks like)')
            print ('              # MARTAS directory')
            print ('              martasdir            :   /home/cobs/MARTAS/')
            print ('              # Define data source (file, db)')
            print ('              source               :   file')
            print ('              # If source = db then define data base credentials created by addcred (MARTAS)')
            print ('              dbcredentials        :   None')
            print ('              # If source = file define the base path')
            print ('              sensorpath           :   /srv/mqtt/')
            print ('              # Notifaction (uses martaslog class, one of email, telegram, mqtt, log) ')
            print ('              notification         :   email')
            print ('              notificationconfig   :   /etc/martas/notification.cfg')
            print ('              # serial communication for switch commands (based on ardcomm.py (MARTAS/app)')
            print ('              serialcfg            :   None')
            print ('              #parameter (all given parameters are checked in the given order, use semicolons for parameter list):')
            print ('              # sensorid; timerange to check; key to check, value, lower or upper bound,statusmessage,resetby,switchcommand(optional)')
            print ('              1  :  DS18B20XX,1800,t1,5,low,average,on,swP:4:1')
            print ('              2  :  DS18B20XY,1800,t1,10,high,median,off,swP:4:0')
            print ('              3  :  DS18B20XZ,600,t2,20,high,max,alarm at date,None')
            print ('              #to be continued...')

            print ('------------------------------------------------------')
            print ('Example:')
            print ('   python threshold.py -m /etc/martas/threshold.cfg')
            sys.exit()
        elif opt in ("-m", "--configfile"):
            configfile = arg
            print ("Getting all parameters from configration file: {}".format(configfile))
            conf = mameth.get_conf(configfile, confdict=conf)
            para = assign_parameterlist(sp.valuenamelist,conf)
        elif opt in ("-D", "--debug"):
            debug = True
        elif opt in ("-T", "--Test"):
            travistestrun = True
            conf = mameth.get_conf(os.path.join('../..', 'conf', 'threshold.cfg'))
            para = assign_parameterlist(sp.valuenamelist,conf)

        # TODO activate in order prevent default values
        #if not configfile or conf == {}:
        #    print (' !! Could not read configuration information - aborting')
        #    sys.exit()

    if debug:
        print ("Configuration dictionary: \n{}".format(conf))
        print ("Parameter dictionary: \n{}".format(para))

    if not (len(para)) > 0:
        print ("No parameters given too be checked - aborting")
        sys.exit()


    try:
        logpath = conf.get('bufferpath')
    except:
        print ("Could not import martas logging routines - check MARTAS directory path")

    # For each parameter
    for i in range(0,1000):
            valuedict = para.get(str(i),{})
            content = ''
            if not valuedict == {}:
                if debug:
                    print ("Checking parameterset {}".format(i))
                data = DataStream()
                testvalue = None
                evaluate = {}

                # Obtain a magpy data stream of the respective data set
                if debug:
                    print ("Accessing data from {} at {}: Sensor {} - Amount: {} sec".format(conf.get('source'),conf.get('bufferpath'),valuedict.get('sensorid'),valuedict.get('timerange') ))

                (data,msg1) = get_data(conf.get('source'), conf.get('bufferpath'), conf.get('dbcredentials'), valuedict.get('sensorid'),valuedict.get('timerange'), debug=debug , startdate=conf.get('startdate') )
                (testvalue,msg2) = get_test_value( data, valuedict.get('key'), valuedict.get('function'), debug=debug) # Returns comparison value(e.g. mean, max etc)
                if not testvalue and travistestrun:
                    print ("Testrun for parameterset {} OK".format(i))
                elif not testvalue and not testvalue == 0.0:
                    print ("Please check your test data set... are bufferfiles existing? Is the sensorid correct?")
                elif is_number(testvalue):
                    (evaluate, msg) = check_threshold(testvalue, valuedict.get('value'), valuedict.get('state'), debug=debug) # Returns statusmessage
                    if evaluate and msg == '':
                        content = interprete_status(valuedict,debug=debug)
                        # Perform switch and added "switch on/off" to content
                        if not valuedict.get('switchcommand') in ['None','none',None,""]:
                            if debug:
                                print ("Found switching command ... eventually will send serial command (if not done already) after checking all other commands")
                            content = '{} - switch: {}'.format(content, valuedict.get('switchcommand'))
                            # remember the switchuing command and only issue it if statusdict is changing
                        if not valuedict.get('executecommand',None) in ['None','none',None,""]:
                            print ("Execute criteria found: running {}".format(valuedict.get('executecommand')))
                            if not os.path.isfile(valuedict.get('executecommand')):
                                print (" - did not find an appropriate script - skipping")
                            else:
                                exe = ['/bin/sh',valuedict.get('executecommand')]
                                if not debug:
                                    import subprocess
                                    subprocess.check_output(exe)
                                else:
                                    print (" - debug selected - otherwise would execute {}".format(exe))
                    elif not msg == '':
                        content =  msg
                    else:
                        content = 'fine'
                else:
                    content = msg1+' - '+msg2


                if conf.get('reportlevel') == 'partial' and content == "fine":
                    content = ""
                    # evetually replace content="fine" with content="" for "partial"
                statinput = '{}-{}-threshold of key {}'.format(hostname, valuedict.get('sensorid'),valuedict.get('key'))
                if statinput in statusdict:
                    if not content in ["fine","",None]:
                        statusdict[statinput] = content
                elif content:
                    statusdict[statinput] = content
                statuskeylist.append(statinput)
                #if content:
                #    statuskeylist.append('Sensor {} and key {}'.format(valuedict.get('sensorid'),valuedict.get('key')))
                #    statusdict['Sensor {} and key {}'.format(valuedict.get('sensorid'),valuedict.get('key'))] = content

                if debug:
                    print ("Finished parameterset {}".format(i))

    #if conf.get('reportlevel') == 'full':
    #    # Get a unique status key list:
    #    statuskeylist = list(dict.fromkeys(statuskeylist))
    #    for elem in statuskeylist:
    #        cont = statusdict.get(elem,'')
    #        if cont == '':
    #            statusdict[elem] = "Everything fine"

    if debug:
        print ("Statusdict: {}".format(statusdict))

    receiver = conf.get('notification')
    cfg = conf.get('notificationconfig')
    logfile = conf.get('logfile')

    if conf.get('https'):
        proxies['https'] = conf.get('https')
    if conf.get('http'):
        proxies['http'] = conf.get('http')

    if travistestrun:
        print ("Skipping send routines in test run - finished successfully")
        sys.exit()

    if debug:
        print ("New notifications would be send to: {} (Config: {})".format(receiver,cfg))
    else:
        martaslog = ml(logfile=logfile,receiver=receiver,proxies=proxies)
        if receiver == 'telegram':
            martaslog.telegram['config'] = cfg
        elif receiver == 'email':
            martaslog.email['config'] = cfg
        elif receiver == 'mqtt':
            martaslog.mqtt['config'] = cfg
        changes = martaslog.msg(statusdict)

    if not len(changes) > 0:
        print ("Nothing to report - threshold check successfully finished")

    for element in changes:
        line = changes.get(element)
        if debug:
            print ("Changes affecting:", element)
        l = line.split('switch:')
        if len(l) == 2:
            print (" ... now dealing with switching serial command:")
            comm = l[1].strip()
            script = os.path.join(conf.get('martasdir'),'app','ardcomm.py')
            pythonpath = sys.executable
            arg1 = "-c {}".format(comm)
            arg2 = "-p {}".format(conf.get('port'))
            arg3 = "-b {}".format(conf.get('baudrate'))
            arg4 = "-a {}".format(conf.get('parity'))
            arg5 = "-y {}".format(conf.get('bytesize'))
            arg6 = "-s {}".format(conf.get('stopbits'))
            arg7 = "-t {}".format(conf.get('timeout'))
            #arg8 = "-e {}".format(conf.get('eol')) # not used so far

            command = "{} {} {} {} {} {} {} {} {}".format(pythonpath,script,arg1, arg2, arg3, arg4, arg5, arg6, arg7) ## Too be checked
            command = "{} {} {}".format(pythonpath,script,arg1)
            if debug:
                print (" ... sending {}".format(command))


            try:
                import subprocess
                p = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True)
                (output, err) = p.communicate()
                mesg = "{}".format(output)
            except subprocess.CalledProcessError:
                mesg = "threshold: sending command didnt work"
            except:
                mesg = "threshold: sending command problem"

            print (mesg)

            print (" ... success")

if __name__ == "__main__":
   main(sys.argv[1:])



#!/usr/bin/env python
"""
MagPy - OPTIMIZING tables and free space
        unblocking version

REQUIREMENTS:
 - magpy
 - sudo apt install percona-toolkit
 - main user (cobs) needs to be able to use sudo without passwd (add to /etc/sudoers)
"""

# Define packgaes to be used (local refers to test environment)
# ------------------------------------------------------------

from magpy.stream import *
import pymysql as mysql
from magpy.core import database
import magpy.opt.cred as mpcred
from martas.core.methods import martaslog as ml
from martas.core import methods as mm
from martas.version import __version__
import socket
import getopt


def optimize(credentials='cobsdb', sqlcred='sql', debug=False):
    """
    DESCRIPTION
        will run an optimization method on all data tables based
        on the percona toolkit
    REQUIREMENTS
        magpy
        sudo apt install percona-toolkit
        main user (cobs) needs to be able to use sudo without passwd (add to /etc/sudoers)
    APPLICATION
        optimize(credentials='cobsdb', sqlcred='sql', debug=True)
    """
    dbpasswd = mpcred.lc(credentials,'passwd')
    dbuser = mpcred.lc(credentials,'user')
    dbhost = mpcred.lc(credentials,'host')
    dbname = mpcred.lc(credentials,'db')

    sqluser = mpcred.lc(sqlcred, 'user')
    sqlpwd = mpcred.lc(sqlcred, 'passwd')

    report = 'optimization failed'

    # Connect to test database
    # ------------------------------------------------------------
    try:
        if debug:
            print("Connecting to DATABASE...")
        db = database.DataBank(host=dbhost, user=dbuser, password=dbpasswd, database=dbname)
    except:
        if debug:
            print("... failed - aborting")
        sys.exit()

    # some general initializations
    cursor = db.db.cursor()
    step1 = True
    step2 = False
    step3 = False
    optimzesql = []
    tables = []

    print ("1. Get initial database information")
    print ("-----------------------------------")
    db.info(destination='stdout',level='full')

    print ("2. Get all tables")
    print ("-----------------------------------")
    tablessql = 'SHOW TABLES'
    try:
        cursor.execute(tablessql)
    except mysql.IntegrityError as message:
        print (message)
        step1 = False
    except mysql.Error as message:
        print (message)
        step1 = False
    except:
        print ('optimize: unknown error')
        step1 = False

    if step1:
        tables = cursor.fetchall()
        tables = [el[0] for el in tables]
        if len(tables) > 0:
            print ('optimize: tables found - continuing')
            step2 = True
            step3 = True
    else:
        print ('optimize: aborting')
        cursor.close()

    if step2 and not debug:
        print ("2. Optimizing tables")
        print ("-----------------------------------")
        for table in tables:
            print (' -> running for {}, user: {}'.format(table, sqluser))
            optbash = 'sudo /usr/bin/pt-online-schema-change --alter "Engine=InnoDB" D={},t={} --user={} --password={} --execute'.format(dbname,table,sqluser,sqlpwd)
            print (optbash)
            try:
                os.system(optbash)
                print ("Done")
            except:
                print ('------------------------------------------')
                print (' --optimize table: unknown error for {}'.format(table))
                print ('------------------------------------------')

    if step3:
        print ("3. Get post-processing information")
        print ("-----------------------------------")
        db.db.commit()
        cursor.close()
        report = db.info(destination='stdout',level='full')

    return report


def main(argv):
    version = __version__
    sn = socket.gethostname().upper()  # servername
    statusmsg = {}
    name = "{}-DBopt".format(sn)
    credentials = ''
    sql = 'sql'
    configpath = ""
    logpath = "/var/log/magpy/tg_db_opt.log"
    conf = {}
    proxies = {}
    telegramcfg = ""
    debug=False

    try:
        opts, args = getopt.getopt(argv,"hc:d:s:l:D",["config=","database=","sql=","logpath=","debug=",])
    except getopt.GetoptError:
        print ('cobs_optimizetables.py -c <config>')
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print ('-------------------------------------')
            print ('Description:')
            print ('-- optimizetables.py will analyse magnetic data --')
            print ('-----------------------------------------------------------------')
            print ('detailed description ..')
            print ('...')
            print ('...')
            print ('-------------------------------------')
            print ('Usage:')
            print ('python optimizetables.py -c <config>')
            print ('-------------------------------------')
            print ('Options:')
            print ('-c (required) : configuration data path')
            print ('-d            : database credentials')
            print ('-s            : sql')
            print ('-l            : logpath')
            print ('-------------------------------------')
            print ('Application:')
            print ('python3 optimzetables.py -d cobsdb -s sql')
            sys.exit()
        elif opt in ("-c", "--config"):
            configpath = os.path.abspath(arg)
        elif opt in ("-d", "--database"):
            credentials = arg
        elif opt in ("-s", "--sql"):
            # define a dayrange for archive jobs - default is 2
            sql = arg
        elif opt in ("-l", "--logpath"):
            logpath = arg
        elif opt in ("-D", "--debug"):
            debug = True

    receiver = ""
    cfg = ""
    if configpath:
        conf = mm.get_conf(configpath)
        basepath = os.path.dirname(conf.get("logging",""))
        if not basepath:
            basepath = os.path.dirname(conf.get("logpath",""))
        logpath = os.path.join(basepath,"optimize.log")
        receiver = conf.get('notification', "telegram")
        cfg = conf.get('notificationcfg', os.path.join(basepath,"..","conf","telegram.cfg"))
        if not credentials:
            credentials = conf.get("dbcredentials")
    print ("Running optimize version", version)
    print ("-------------------------------")
    report = optimize(credentials=credentials, sqlcred=sql, debug=debug)

    # Send out a new status
    statusmsg[name] = report
    print("Status", statusmsg[name])

    if conf.get('https'):
        proxies['https'] = conf.get('https')
    if conf.get('http'):
        proxies['http'] = conf.get('http')

    if report and cfg and not debug:
        martaslog = ml(logfile=logpath, receiver=receiver, proxies=proxies)
        if receiver == 'telegram':
            martaslog.telegram['config'] = cfg
        elif receiver == 'email':
            martaslog.email['config'] = cfg
        martaslog.msg(statusmsg)
    else:
        print("Status", statusmsg[name])


if __name__ == "__main__":
   main(sys.argv[1:])



from __future__ import print_function
from __future__ import absolute_import

# ###################################################################
# Import packages
# ###################################################################

from magpy.stream import DataStream, KEYLIST, NUMKEYLIST, subtractStreams
import struct
from datetime import datetime 
import json
import base64
import binascii

def datetime2array(t):
        return [t.year,t.month,t.day,t.hour,t.minute,t.second,t.microsecond]

## LORA-ZAMG - protocol
##
class lorawanserver_status(object):
    """
       ********************************
       *                              *
       *   Gateway Status Meldungen   *
       *                              *
       ********************************

       Debugging: adding once after the other

       mosquitto_sub -h 138.22.165.37 -t "#" -v |grep stats
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:06:46Z","rxPacketsReceived":1,"rxPacketsReceivedOK":1,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:07:16Z","rxPacketsReceived":1,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:07:46Z","rxPacketsReceived":0,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:08:16Z","rxPacketsReceived":0,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:08:46Z","rxPacketsReceived":0,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:09:16Z","rxPacketsReceived":0,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:09:46Z","rxPacketsReceived":0,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:10:16Z","rxPacketsReceived":0,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:10:46Z","rxPacketsReceived":0,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:11:16Z","rxPacketsReceived":2,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:11:46Z","rxPacketsReceived":2,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:12:16Z","rxPacketsReceived":0,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:12:46Z","rxPacketsReceived":0,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:13:16Z","rxPacketsReceived":1,"rxPacketsReceivedOK":1,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:13:46Z","rxPacketsReceived":2,"rxPacketsReceivedOK":2,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:14:16Z","rxPacketsReceived":0,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:14:46Z","rxPacketsReceived":1,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:15:16Z","rxPacketsReceived":0,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:15:46Z","rxPacketsReceived":0,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:16:16Z","rxPacketsReceived":0,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:16:46Z","rxPacketsReceived":0,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:17:16Z","rxPacketsReceived":1,"rxPacketsReceivedOK":1,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}jj}
       gateway/00800000a0001285/stats {"mac":"00800000a0001285","time":"2018-12-14T16:17:46Z","rxPacketsReceived":0,"rxPacketsReceivedOK":0,"txPacketsReceived":0,"txPacketsEmitted":0,"customData":{"ip":"138.22.165.40"}}

    """
    
    def __init__(self):
        """
        """
        print ("  -> Initializing loraWAN server routines status 'stats' ...")
        self.topicidentifier = {'startswith':'gateway','endswith':'stats'}
        self.stats_datakeytranslator = {
                'rxPacketsReceived':['rxPR1','int'], 
                'rxPacketsReceivedOK':['rxPRO','int'], 
                'txPacketsReceived':['txPR','int'], 
                'txPacketsEmitted':['txPE','int'],
                'customData':[{'ip':'ip'}]
                }
        self.identifier = {}
        self.headdict = {}
        self.headstream = {}
    
    def GetPayload(self, payload, topic):
        # NO: only with paho: loradict = json.loads(payload)
        # JF alread dictionary if is json formated: type(payload) => dict
        # JF only at comandline with python *.py: loradict = payload
        loradict = json.loads(payload)
        newpayload, sensorid, headline, header =  self.loradict2datastruct(loradict)
        return newpayload, sensorid, headline, header, self.identifier
    
    def loradict2datastruct(self, loradict):
        # JF NO string: stats_datakeytranslator = { 'rxPacketsReceived':['rxPR1','int'], 'rxPacketsReceivedOK':['rxPRO','int'], 'txPacketsReceived':['txPR','int'], 'txPacketsEmitted':['txPE','int'], 'customData':['ip','ipv4']}
            stats_datakeytranslator = { 'rxPacketsReceived':['rxPR1','int'], 'rxPacketsReceivedOK':['rxPRO','int'], 'txPacketsReceived':['txPR','int'], 'txPacketsEmitted':['txPE','int']}
            # customdatadict = loradict.get('customData')[0]
            # JF NO is no object => removed: [0]
            customdatadict = loradict.get('customData')
            mac = str(loradict.get('mac'))
            header = {}
            header['mac'] = loradict.get('mac')
            header['ip']  = customdatadict.get('ip')
            # sensorid = header['SensorName'][:5].replace('-','') + '_' + header['SensorSerialNum'] + '_0001'
            # header['SensorID'] = sensorid
            sensorid = "Gateway_" + mac + "_" + self.topicidentifier['endswith'] + '_1000'
            header['SensorID'] = sensorid
            rxPR   = loradict.get('rxPacketsReceived')
            rxPROK = loradict.get('rxPacketsReceivedOK')
            txPR   = loradict.get('txPacketsReceived')
            txPE   = loradict.get('txPacketsEmitted')
            ip = '.'.join([i.zfill(3) for i in customdatadict.get('ip').split('.')])
            dataline      = {"rxPacketsReceived":rxPR,"rxPacketsReceivedOK":rxPROK,"txPacketsReceived":txPR,"txPacketsEmitted":txPE}
            """
            """
            # JF NO string: datadict = {"rxPacketsReceived":rxPR, "rxPacketsReceivedOK":rxPROK, "txPacketsReceived":txPR, "txPacketsEmitted":txPE, customData':["ip",ip]}
            datadict = {"rxPacketsReceived":rxPR, "rxPacketsReceivedOK":rxPROK, "txPacketsReceived":txPR, "txPacketsEmitted":txPE]}
            keylist, elemlist, unitlist, multilist = [],[],[],[]
            if not loradict.get('time','') == '':
                """ 2018-12-14T16:10:16Z """
                time = datetime.strptime(loradict.get('time'),"%Y-%m-%dT%H:%M:%SZ")
            elif not loradict.get('version','') == '':
                time = datetime.strptime(loradict.get('DatumTime'),"%Y-%m-%dT%H:%M:%S.%fZ")
            else:
                time = datetime.now(timezone.utc).replace(tzinfo=None)
            
            datalst = datetime2array(time)
            packstr = '6hL'
            for elem in datadict:
                if elem in stats_datakeytranslator:
                    key = stats_datakeytranslator[elem][0]
                    unit = stats_datakeytranslator[elem][1]
                    keylist.append(key)
                    elemlist.append(elem)
                    unitlist.append(unit)
                    multilist.append(1000)
                    packstr += "l"
                    if type(datadict[elem]) == int:
                        datalst.append(int(datadict[elem]*1000))   
                    else:
                        datalst.append(str(datadict[elem]))   
                # print (elem, datadict[elem])
            
            datalst = [str(elem) for elem in datalst]
            dataline =','.join(datalst)
            self.identifier[sensorid+':packingcode'] = packstr
            self.identifier[sensorid+':keylist'] = keylist
            self.identifier[sensorid+':elemlist'] = elemlist
            self.identifier[sensorid+':unitlist'] = unitlist
            self.identifier[sensorid+':multilist'] = multilist
            
            def identifier2line(dic, sensorid):
                p1 = dic.get(sensorid+':packingcode')
                p2 = dic.get(sensorid+':keylist')
                p3 = dic.get(sensorid+':elemlist')
                p4 = dic.get(sensorid+':unitlist')
                p5 = dic.get(sensorid+':multilist')
                p5 = [str(elem) for elem in p5]
                size = struct.calcsize(p1)
                line = "# MagPyBin {} [{}] [{}] [{}] [{}] {} {}".format(sensorid,','.join(p2),','.join(p3),','.join(p4),','.join(p5),p1,size)
                return line
            
            headline = identifier2line(self.identifier, sensorid)
            print ('dataline', dataline)
            print ('sensorid', sensorid)
            print ('headline', headline)
            print ('header', header)
            print ("success")
            
            return dataline, sensorid, headline, header




# ###################################################################
# Import packages
# ###################################################################

import time
import struct # for binary representation
import socket # for hostname identification
import string # for ascii selection
from datetime import datetime, timezone
from twisted.python import log
from martas.core import methods as mm
import subprocess


try:
    import board
    import adafruit_mmc56x3
except:
    log.msg("   mmc5603 i2c requires the following packages: board, adafruit-blinka and adafruit-circuitpython-mmc56x3")

    ## I2C MMC5603 protocol
## -----------------------

class MMC5603I2CProtocol(object):
    """
    Protocol to read MMC5603 data attached to I2C pins of i.e. an raspberry
    This protocol makes use of the Adafruit library.
    pip3 install adafruit-circuitpython-mmc56x3

    """

    def __init__(self, client, sensordict, confdict):
        self.client = client
        self.sensordict = sensordict
        self.confdict = confdict
        self.count = 0  ## counter for sending header information
        self.sensor = sensordict.get('sensorid')
        self.hostname = socket.gethostname()
        self.printable = set(string.printable)
        self.datalst = []
        self.datacnt = 0
        self.metacnt = 10
        self.counter = {}
        self.initi2c = True

        # extract data from configuration dictionary
        debugtest = confdict.get('debug')

        # QOS
        self.qos=int(confdict.get('mqttqos',0))
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)


        # switch on debug mode
        self.debug = False
        if debugtest in ['true','True','TRUE']:
            log.msg('DEBUG - {}: Debug mode activated.'.format(self.sensordict.get('protocol')))
            self.debug = True    # prints many test messages
        else:
            log.msg('  -> Debug mode = {}'.format(debugtest))

        # connect to i2c
        self.conn = self.init_sensor()


    def datetime2array(self,t):
        return [t.year,t.month,t.day,t.hour,t.minute,t.second,t.microsecond]

    def init_sensor(self):
        #log.msg("connecting to I2C ...")
        i2c = board.I2C()  # uses board.SCL and board.SDA
        # i2c = board.STEMMA_I2C()  # For using the built-in STEMMA QT connector on a microcontroller
        sensor = adafruit_mmc56x3.MMC5603(i2c)
        time.sleep(0.5)
        sensor.reset()  # in Hz, from 1-255 or 1000
        time.sleep(0.5)
        sensor.set_reset()  # in Hz, from 1-255 or 1000
        #sensor.data_rate = 10  # in Hz, from 1-255 or 1000
        #sensor.continuous_mode = True
        return sensor


    def processMMC5603Data(self, sensorid, meta, data):
        """Convert raw ADC counts into SI units as per datasheets"""
        currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
        outdate = datetime.strftime(currenttime, "%Y-%m-%d")
        filename = outdate

        datearray = self.datetime2array(currenttime)

        packcode = '6hL'

        values = []
        multiplier = []
        for dat in data:
            try:
                values.append(float(dat))
                datearray.append(int(float(dat)*10000))
                packcode = packcode + 'q'
                multiplier.append(10000)
            except:
                log.msg('{} protocol: Error while appending data to file (non-float?): {}'.format(self.sensordict.get('protocol'),dat) )

        try:
            data_bin = struct.pack('<'+packcode,*datearray) #little endian
        except:
            log.msg('{} protocol: Error while packing binary data'.format(self.sensordict.get('protocol')))
            pass

        key = '['+str(meta.get('SensorKeys')).replace("'","").strip()+']'
        ele = '['+str(meta.get('SensorElements')).replace("'","").strip()+']'
        unit = '['+str(meta.get('SensorUnits')).replace("'","").strip()+']'
        multplier = str(multiplier).replace(" ","")
        # Correct some common old problem
        unit = unit.replace('deg C', 'degC')

        header = "# MagPyBin %s %s %s %s %s %s %d" % (sensorid, key, ele, unit, multplier, packcode, struct.calcsize(packcode))

        if not self.confdict.get('bufferdirectory','') == '':
            mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, filename, data_bin, header)

        return ','.join(list(map(str,datearray))), header

    def get_mmc5603_data(self,sensor, debug=False):
        mag_x, mag_y, mag_z = sensor.magnetic
        temp = sensor.temperature
        if debug:
             print(f"X:{mag_x:10.3f}, Y:{mag_y:10.3f}, Z:{mag_z:10.3f} uT, T:{temp:10.3f} degC")
        data = [float(mag_x)*1000., float(mag_y)*1000., float(mag_z)*1000., temp]
        return data

    def calibrate_mmc5603_data(self, data, offsets=None, scales=None, debug=False):
        if not offsets:
            offsets=[0,0,0,0]
        if not scales:
            scales=[1,1,1,1]
        for i, el in enumerate(data):
            data[i] = el+offsets[i]
            data[i] = data[i]*scales[i]
        return data

    def define_mmc5603_meta(self):
        meta = {}
        meta['SensorKeys'] = 'x,y,z,t1'
        meta['SensorElements'] = 'X,Y,Z,Ts'
        meta['SensorUnits'] = 'nT,nT,nT,C'
        return meta

    def sendRequest(self):

        if not self.conn:
            self.conn = self.init_sensor()
            # or put everything in a try except and reconnect on except
        # connect to i2c
        data = self.get_mmc5603_data(self.conn, debug=self.debug)
        data = self.calibrate_mmc5603_data(data, offsets=[0,0,550000,9])
        meta = self.define_mmc5603_meta()
        self.publish_data(data,meta)

    def publish_data(self, data, meta):

        if len(data) > 0:
            topic = self.confdict.get('station') + '/' + self.sensor
            pdata, head = self.processMMC5603Data(self.sensor, meta, data)

            senddata = False
            try:
                coll = int(self.sensordict.get('stack'))
            except:
                coll = 0

            if coll > 1:
                self.metacnt = 1 # send meta data with every block
                if self.datacnt < coll:
                    self.datalst.append(pdata)
                    self.datacnt += 1
                else:
                    senddata = True
                    pdata = ';'.join(self.datalst)
                    self.datalst = []
                    self.datacnt = 0
            else:
                senddata = True

            if senddata:
                self.client.publish(topic+"/data", pdata, qos=self.qos)
                # If multiple sensors are connected, self.count needs to be a dict (with topic)
                # Initialize counter for each topic
                if self.counter.get(self.sensor,'') == '':
                    self.counter[self.sensor] = 0
                cnt = self.counter.get(self.sensor)

                if cnt == 0:
                    ## 'Add' is a string containing dict info like:
                    ## SensorID:ENV05_2_0001,StationID:wic, PierID:xxx,SensorGroup:environment,...
                    add = "SensorID:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDescription:{},DataTimeProtocol:{}".format( self.sensordict.get('sensorid',''),self.confdict.get('station',''),self.sensordict.get('pierid',''),self.sensordict.get('protocol',''),self.sensordict.get('sensorgroup',''),self.sensordict.get('sensordesc',''),self.sensordict.get('ptime','') )
                    self.client.publish(topic+"/dict", add, qos=self.qos)
                    self.client.publish(topic+"/meta", head, qos=self.qos)

                cnt += 1
                #self.count += 1
                if cnt >= self.metacnt:
                    cnt = 0

                # update counter in dict
                self.counter[self.sensor] = cnt



# ###################################################################
# Import packages
# ###################################################################

import struct # for binary representation
import socket # for hostname identification
import string # for ascii selection
import sys
from datetime import datetime, timezone
from twisted.protocols.basic import LineReceiver
from twisted.python import log
from martas.core import methods as mm
import numpy as np

## GEM -GP20S3 protocol
##
class GP20S3Protocol(LineReceiver):
    """
    Protocol to read GP20S3 data
    This protocol defines the individual sensor related read process.
    It is used to dipatch url links containing specific data.
    Sensor specific coding is contained in method "processData".

    Treatment of GPS and NTP time:
    The protocol determines two times: 1) GPS time provided by the data logger
    2) NTP time of the PC when data is received by serial comm.
    Primary time is selected in sensor.cfg (either NTP or GPS)
    Timediff between GPS and NTP is recorded, and a 1000 line median average is send
    within the DataDictionary (DataNTPTimeDelay).
    Besides, a waring message is generated if NTP and GPS times differ by more than 3 seconds.
    For this comparison (and only here) a typical correction (delay) value is applied to NTP.
    e.g. 6.2 sec for POS1.
    """
    def __init__(self, client, sensordict, confdict):
        """
        'client' could be used to switch between different publishing protocols
                 (e.g. MQTT or WS-MCU gateway factory) to publish events
        'sensordict' contains a dictionary with all sensor relevant data (sensors.cfg)
        'confdict' contains a dictionary with general configuration parameters (martas.cfg)
        """
        self.client = client
        self.sensordict = sensordict
        self.confdict = confdict
        self.count = 0  ## counter for sending header information
        self.sensor = sensordict.get('sensorid')
        self.hostname = socket.gethostname()
        self.printable = set(string.printable)
        #log.msg("  -> Sensor: {}".format(self.sensor))
        self.datalst = []
        self.datacnt = 0
        self.metacnt = 10
        self.errorcnt = {'time':0}

        self.delaylist = []  # delaylist contains up to 1000 diffs between gps and ntp
                             # the median of this values is used for ntp timedelay
        self.timedelay = 0.0
        self.timethreshold = 3 # secs - waring if timedifference is larger the 3 seconds
        self.validitythresholds = {"intensity":[20000000,100000000]} # valid ranges for intensities

        # QOS
        self.qos=int(confdict.get('mqttqos',0))
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)

        # Debug mode
        debugtest = confdict.get('debug')
        #self.debug = False
        #self.debug = True
        if debugtest == 'True':
            log.msg('DEBUG - {}: Debug mode activated.'.format(self.sensordict.get('protocol')))
            self.debug = True    # prints many test messages
        else:
            log.msg('  -> Debug mode = {}'.format(debugtest))
        # PYTHON version
        self.pvers = sys.version_info[0]


    def connectionMade(self):
        log.msg('  -> {} connected.'.format(self.sensor))

    def connectionLost(self, reason):
        log.msg('  -> {} lost.'.format(self.sensor))

    def validity_check(data2check, thresholds=[],debug=False):
        # call by validity_check(intensity1, self.validitythresholds.get('intensity',[]))
        valid = True
        if len(thresholds) > 1:
           if data2check < thresholds[0] or data2check > thresholds[1]:
              valid = False
        return valid

    def processData(self, data):
        """ GP20S3 data """
        """
        Data looks like--- (with GPS lines every minute):
        -- vertical sensor - Old software
        3,3,12.00 111 field1 field2 field3
        3,3,12.00 111 field1 field2 field3
        GPS 16.00 111 field1 field2 field3
        -- horizontal sensor - New software
        time 111 field1 field2 field3                                            (every sec or faster)
        $$$                                                         (every hour, preceeds status line)
        10071506 A 13 250 492 496 329 150 1023 39 39 39 30 29 30 YYYyyyEEENNN 148 149 117 (every hour)
        time 111 field1 field2 field3                                            (every sec or faster)
        time 111 field1 field2 field3                                            (every sec or faster)
        """

        currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
        outdate = datetime.strftime(currenttime, "%Y-%m-%d")
        actualtime = datetime.strftime(currenttime, "%Y-%m-%dT%H:%M:%S.%f")
        outtime = datetime.strftime(currenttime, "%H:%M:%S")
        timestamp = datetime.strftime(currenttime, "%Y-%m-%d %H:%M:%S.%f")
        filename = outdate
        sensorid = self.sensor
        headerlinecoming = False
        datearray = []
        headarray = []
        statusname = "Status_123_0001"

        sensororientation = self.sensor.split('_')[0].replace(self.sensordict.get('protocol'),'')
        if len(sensororientation) > 1:
            sens1 = sensororientation
            sens2 = sensororientation[0]
            sens3 = sensororientation[1]
        else:
            sens1 = 'TA'
            sens2 = 'B'
            sens3 = 'TB'
        celem =  '[{},{},{},{}{},{}{},{}{},None]'.format(sens1,sens2,sens3, sens3,sens1, sens3,sens2, sens2,sens1)
        packcode = '6hLQQQqqq6hL'
        header = "# MagPyBin %s %s %s %s %s %s %d" % (self.sensor, '[x,y,z,dx,dy,dz,sectime]', celem, '[pT,pT,pT,pT,pT,pT,None]', '[1000,1000,1000,1000,1000,1000,1]', packcode, struct.calcsize('<'+packcode))

        try:
            # Extract data
            data_array = data
            if len(data_array) == 5:
                intensity1 = float(data_array[2])
                intensity2 = float(data_array[3])
                intensity3 = float(data_array[4])
                grad1 = intensity3-intensity1
                grad2 = intensity3-intensity2
                grad3 = intensity2-intensity1
                try:
                    gpstime = float(data[0]) # will fail for old dataformat -> NTP
                    if gpstime > 235900.0: # use date of last day if gpstime > 235900 to prevent next day date for 235959 gps when pctime already is on next day
                        cdate = dateprev
                    else:
                        cdate = outdate
                        dateprev = outdate
                    try:
                        internal_t = datetime.strptime(cdate+'T'+data[0], "%Y-%m-%dT%H%M%S.%f")
                    except:
                        internal_t = datetime.strptime(cdate+'T'+data[0], "%Y-%m-%dT%H%M%S")
                    internal_time = datetime.strftime(internal_t, "%Y-%m-%d %H:%M:%S.%f")
                except:
                    internal_time = timestamp #datetime.strftime(datetime.now(timezone.utc).replace(tzinfo=None), "%Y-%m-%d %H:%M:%S.%f")

            elif len(data_array) == 19:
                """
                        10071506 A 13 250 492 496 329 150 1023 39 39 39 30 29 30 YYYyyyEEENNN 148 149 117

			<GPS> day/month/year/hour A - locked, V unlocked
			<13> Console outside air temperature (13C)
			<250> Battery voltage (25.0V)
			<492> +5V supply voltage (4.92V)
			<496> -5V supply voltage (-4.96)
			<3.3V> +3.3V supply voltage (3.3V)
			<15.0> silver box power supply (15.0V)
			<1023> OCXO internal trimpot adjustment level, automatically adjusted via GPS
			<39> Sensor 1 temperature in C
			<39>  Sensor 2 temperature in C
			<39> Sensor 3 temperature in C
			<30> Light current sensor 1 (3.0uA)
			<29> Light current sensor 2 (2.9uA)
			<30> Light current sensor 3 (3.0uA)
			<YYY>  Sensor 1, sensor 2 sensor 3 lock status Y- locked, N - unlocked
			<yyy>  Sensor 1 heater status, sensor 2 heater status, sensor 3 heater status y-on, n-off
			<EEE> Sensor 1 heater, sensor 2 heater, sensor 3 heater E-enabled, D-disabled (used for over heat protection)
			<NNN> RF sensor 1, RF sensor 2, RF sensor 3, N -on, F-off
			<148> Sensor 1 RF dc voltage (14.8V)
			<149> Sensor 2 RF dc voltage (14.9V)
			<117> Sensor 3 RF dc voltage (11.7V)
                """
                headerlinecoming = True

                try:
                    gpstime = str(data_array[0])
                    internal_t = datetime.strptime(gpstime, "%d%m%y%H")
                    gpstimestamp = datetime.strftime(internal_t, "%Y-%m-%d %H:%M:%S.%f")
                except:
                    gpstimestamp = timestamp
                internal_time = gpstimestamp

                gpstatus = data_array[1]			# str1
                telec = int(data_array[2])			# t2
                Vbat = float(data_array[3])/10.			# f
                Vsup1 = float(data_array[4])/100.		# var4
                Vsup2 = float(data_array[5])/100.		# var5
                Vlow = float(data_array[6])/100.		# t1
                PowerSup = float(data_array[7])/10.		# df
                level = data_array[8]				# str3
                tsens1 = int(data_array[9])			# x
                tsens2 = int(data_array[10])			# y
                tsens3 = int(data_array[11])			# z
                lightcurrent1 = float(data_array[12])/10.	# dx
                lightcurrent2 = float(data_array[13])/10.	# dy
                lightcurrent3 = float(data_array[14])/10.	# dz
                statusstring = data_array[15]			# str2
                Vsens1 = float(data_array[16])/10.		# var1
                Vsens2 = float(data_array[17])/10.		# var2
                Vsens3 = float(data_array[18])/10.		# var3

            elif len(data_array) == 1 and data_array[0] == '$$$':
                return "","",""
            else:
                log.msg('{} protocol: data line could not be interpreted: ({}) of length {}'.format(self.sensordict.get('protocol'),data, len(data_array)))
        except:
            log.err('{} protocol: Data formatting error. Data looks like: {}'.format(self.sensordict.get('protocol'),data))

        try:
            # Analyze time difference between GSM internal time and utc from PC
            timelist = sorted([internal_t,currenttime])
            timediff = timelist[1]-timelist[0]
            delta = timediff.total_seconds()
            if not delta in [0.0, float('NAN'), None]:
                self.delaylist.append(timediff.total_seconds())
                if len(self.delaylist) > 600:
                    self.delaylist = self.delaylist[-600:]
            if len(self.delaylist) > 100:
                try:
                    self.timedelay = np.abs(np.median(np.asarray(self.delaylist)))
                except:
                    self.timedelay = 0.0
            if self.timedelay > self.timethreshold:
                self.errorcnt['time'] +=1
                if self.errorcnt.get('time') < 2:
                    log.msg("{} protocol: large time difference observed for {}: {} sec".format(self.sensordict.get('protocol'), sensorid, self.timedelay))
                if self.errorcnt.get('time') > 1000:
                    self.errorcnt['time'] = 1000
            else:
                self.errorcnt['time'] = 0
        except:
            pass

        if self.sensordict.get('ptime','') in ['NTP','ntp']:
            secondtime = internal_time
            maintime = timestamp
        else:
            maintime = internal_time
            secondtime = timestamp

        if not headerlinecoming:
            try:
                ## GP20S3 provides info on whether the GPS reading is OK  - use it
                # Check data validity
                #if validity_check([intensity1,intensity2,intensity3], thresholds,debug=False):

                # extract time data
                datearray = mm.time_to_array(maintime)
                try:
                    datearray.append(int(intensity1*1000.))
                    datearray.append(int(intensity2*1000.))
                    datearray.append(int(intensity3*1000.))
                    datearray.append(int(grad1*1000.))
                    datearray.append(int(grad2*1000.))
                    datearray.append(int(grad3*1000.))
                    internalarray = mm.time_to_array(secondtime)
                    datearray.extend(internalarray)
                    data_bin = struct.pack('<'+packcode,*datearray)
                except:
                    log.msg('{} protocol: Error while packing binary data'.format(self.sensordict.get('protocol')))

                if not self.confdict.get('bufferdirectory','') == '':
                    mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, filename, data_bin, header)
                #else:
                #   datearray = []
            except:
              log.msg('{} protocol: Error with binary save routine'.format(self.sensordict.get('protocol')))

        if headerlinecoming:
            if self.debug:
                print (" now writing header info")
            headpackcode = '6hL15ls12s4s' #'6hLlllllllllllllllsss'
            statusname = "Status_123_0001"
            try:
                # extract time data
                headarray = mm.time_to_array(maintime)
                try:
                    headarray.append(int(tsens1))			# x
                    headarray.append(int(tsens2))			# y
                    headarray.append(int(tsens3))			# z
                    headarray.append(int(Vbat*10.))			# f
                    headarray.append(int(Vlow*100.))		# t1
                    headarray.append(int(telec))			# t2
                    headarray.append(int(lightcurrent1*10.))	# dx
                    headarray.append(int(lightcurrent2*10.))	# dy
                    headarray.append(int(lightcurrent3*10.))	# dz
                    headarray.append(int(PowerSup*10.))		# df
                    headarray.append(int(Vsens1*10.))		# var1
                    headarray.append(int(Vsens2*10.))		# var2
                    headarray.append(int(Vsens3*10.))		# var3
                    headarray.append(int(Vsup1*100.))		# var4
                    headarray.append(int(Vsup2*100.))		# var5
                    headarray.append(gpstatus)			# str1
                    headarray.append(statusstring)			# str2
                    headarray.append(level)				# str3

                    data_bin = struct.pack('<'+headpackcode,*headarray)
                    statuslst = self.sensor.split('_')
                    if self.debug:
                        print ("Headerdata has been packed")
                    if len(statuslst) == 3:
                        statusname = '_'.join([statuslst[0]+'status',statuslst[1],statuslst[2]])
                    headheader = "# MagPyBin %s %s %s %s %s %s %d" % (statusname, '[x,y,z,f,t1,t2,dx,dy,dz,df,var1,var2,var3,var4,var5,str1,str2,str3]', '[Ts1,Ts2,Ts3,Vbat,V3,Tel,L1,L2,L3,Vps,V1,V2,V3,V5p,V5n,GPSstat,Status,OCXO]', '[degC,degC,degC,V,V,degC,A,A,A,V,V,V,V,V,V,None,None,None]', '[1,1,1,10,100,1,10,10,10,10,10,10,10,100,100,1,1,1]', headpackcode, struct.calcsize('<'+headpackcode))
                    if self.debug:
                        print ("Header looks like: {} ".format(headheader))
                        print ("Writing to file: {}, {}, {}".format(statusname,filename,headheader))
                    if not self.confdict.get('bufferdirectory','') == '':
                        mm.data_to_file(self.confdict.get('bufferdirectory'), statusname, filename, data_bin, headheader)
                except:
                    log.msg('GSMP20 - Protocol: Error while packing binary data')
            except:
                pass

        if len(datearray) > 0:
            topic = self.confdict.get('station') + '/' + self.sensordict.get('sensorid')
            return ','.join(list(map(str,datearray))), header, topic
        elif len(headarray) > 0:
            topic = self.confdict.get('station') + '/' + statusname
            return ','.join(list(map(str,headarray))), headheader, topic
        else:
            return '', '', ''


    def lineReceived(self, line):

        # Defaulttopic
        topic = self.confdict.get('station') + '/' + self.sensordict.get('sensorid')
        if self.pvers > 2:
            line=line.decode('ascii')
        # extract only ascii characters
        line = ''.join(filter(lambda x: x in string.printable, str(line)))
        line = line.replace("b'","").replace("'","")
        ok = True
        try:
            splitline = line.split()
            data, head, topic = self.processData(splitline)
        except:
            data = ''
            print('{}: Data seems not be correct data: Looks like {}'.format(self.sensordict.get('protocol'),line))
            ok = False

        if data =="":
            ok = False

        if ok:
            senddata = False
            coll = int(self.sensordict.get('stack'))
            if coll > 1:
                self.metacnt = 1 # send meta data with every block
                if self.datacnt < coll:
                    self.datalst.append(data)
                    self.datacnt += 1
                else:
                    senddata = True
                    data = ';'.join(self.datalst)
                    self.datalst = []
                    self.datacnt = 0
            else:
                senddata = True

            if senddata:
                self.client.publish(topic+"/data", data, qos=self.qos)
                if self.count == 0:
                    add = "SensorID:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDecription:{},DataTimeProtocol:{}".format( self.sensordict.get('sensorid',''),self.confdict.get('station',''),self.sensordict.get('pierid',''),self.sensordict.get('protocol',''),self.sensordict.get('sensorgroup',''),self.sensordict.get('sensordesc',''),self.sensordict.get('ptime','') )
                    self.client.publish(topic+"/dict", add, qos=self.qos)
                    self.client.publish(topic+"/meta", head, qos=self.qos)
                self.count += 1
                if self.count >= self.metacnt:
                    self.count = 0
